#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
| `@purpose`: Empty path for pylink.  
| `@date`: Created on Sat May 1 15:12:38 2019  
| `@author`: Semeon Risom  
| `@email`: semeon.risom@gmail.com  
| `@url`: https://semeon.io/d/mdl
"""

# core
from pdb import set_trace as breakpoint
import os
import sys

__name__ = 'pylink'
id = 'pylink'

class EyeLinkCustomDisplay():
	pass

# relative paths
sys.path.append(os.path.dirname(os.path.realpath(__file__)))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| @purpose: Default settings for mdl.r33.Processing.  
| @date: Created on Sat May 1 15:12:38 2019
| @author: Semeon Risom
| @email: semeon.risom@gmail.com
| @url: https://semeon.io/d/R33-analysis
"""
# available functions
__all__ = ['Settings','config']

# required external library
__required__ = ['datetime','distutils','importlib']

# global
from pdb import set_trace as breakpoint
import os
import re

# local libraries
if __name__ == '__main__':
	from .. import settings

config={
    #!!!----path
    'path': {
        'home': os.path.abspath(os.getcwd()),
        'root': os.path.abspath(os.getcwd()),
        'r': os.path.abspath(os.getcwd() + 'dist/_R/'),
        'output': os.path.abspath(os.getcwd()+ 'dist/output'),
        'processed': os.path.abspath(__file__+ 'dist/output/data/process/'),
        'summary': os.path.abspath(__file__+ 'dist/output/data/R33-dotprobe-js.csv'),
    },
    #!!!----multiprocessing
    'cores': 7, #number of cpu cores to use for multiprocessing (all subjects only)
    #!!!----style
    'style':{
        'seaborn':'ticks'
    },
    #!!!----metadata
    'metadata':{
        #----analysis
        'subjects': {},
        'is': {}, #bool
        'var': {}, #raw variable name
        'short':{}, #short-hand name of variable
        'long':{}, #long name of variable
        'def': {}, #variable definition
        'cite': {}, #variable citation
        'url': {}, #url
        'img': {}, #image example
        'events': {'fixation':1500,'stimulus':{'iaps':4500,'pofa':3000}},
        'redcap': {'token':'D04484634409375EA8CC34F5B71BC14A'},
        #----lab equipment
        'lab':{
            'monitor.cm': [52, 29.5], #screensize (cm)
            'resolution.px': [1920, 1080], #screensize (px)
            'lab_monitor_device': 'Dell UltraSharp U2414H',
            'eyelink_webcam': 'Eyelink 1000 Plus',
            'lab_monitor_device': 'Dell UltraSharp U2414H',
            'lab_webcam': 'Logitech C922 Pro Stream Webcam',
            'distance': 615.0, #distance from screen (mm)
        },
        #----articles, notes
        'articles':[
            'One algorithm to rule them all? An evaluation and discussion of ten eye movement event-detection algorithms',
            'Comparison of eye movement filters used in HCI'
        ]
    },
    #!!!----processing (for processing.py)
    'processing':{
        'task': 'gRT',
        'type': 'eyetracking',
        'single_subject': False,
        'single_trial': False,
    },
    #!!!----preprocessing (for processing.py)
    'preprocessing':{
        'remove_missing': False, # remove missing data
        'remove_bounds': False, # remove samples out of bounds (i.e. gx=(2255, -322)
        'remove_spikes': False, # remove one-sample spikes
        'spike_delta': 50,
    },
    #!!!----filtering parameters
    'filter':{
        'f_b':{'N':2,'Wn':0.2,'btype':'low'},
        'f_g':{'sigma':2},
        'f_m':{'size':5},
        'f_sg':{'window':11,'order':3}, #default window=11,order=3
        'f_a':{'weights':([1.0, 2.0, 3.0, 2.0, 1.0])},
    },
    #!!!---classifcation parameters
    'classify':{
        'is_classify': True,
        'classify_eyelink_data': True, # if data is from eyelink, so we self classify fixations (True) or use original (False)
        'ctype': 'hmm', #if self classifying fixations, what classification technique should we use #either hmm, simple, ivt, or idt
        #simple parameters
        'missing': 0.0, #value to be used for missing data (simple)
        'maxdist': 360, #maximal inter sample distance in pixels (simple)
        'mindur': 200, #minimal duration of a fixation in milliseconds
        #ivt parameters
        'v_th': 5, #Velocity threshold in px/sec (ivt; default 20)
        #idt parameters
        'dr_th': 50, #Fixation duration threshold in px/msec (idt; default 100)
        'di_th': 60, #Dispersion threshold in px (idt; default 20)
        'filters':[['SavitzkyGolay','sg']]
    }
}

class Settings():
    """Default settings for mdl.r33.Processing"""

	# default configuration parameters
    config = config

    def __init__(self, isLibrary=False):
        """
		Initiate the mdl.r33.Settings module.

		Parameters
		----------
		is_library : :class:`bool` or `list`
			Check if required libraries are available. Default `False`.
		"""
        #check libraries
        if isLibrary:
            settings.library(__required__)

    @classmethod
    def definitions(cls, config):
        """
        Store definitions.

        Parameters
        ----------
        message : :class:`str`
            Log message.
        source : :class:`str`
            Origin of call. Either debug or timestamp.

        Returns
        -------
        config : :class:`dict`
            Returned dictionary

        Examples
        --------
        CESD Group
            in-text: m_['short']['cesd_group'] = 'CESD Group'
            title: m_['long']['cesd_group'] = 'CESD Group'
            definition: m_['def']['cesd_group'] "a binary measure of CESD score (between subjects; 'Low' (&lt16) and 'High' (≥16))"
        """
        #----prepare dictionary
        m_ = {
            'var':{},
            'short':{},
            'long':{},
            'def':{},
            'url':{},
            'cite':{},
            'img':{},
        }
        #----methods
        # task design
        name = 'task_design'
        m_['def'][name] = "Treatment stimuli include pairs of faces, each from a different actor and each expressing sad and \
        neutral emotions, from the Pictures of Facial Affect (POFA; [27]) collection and dysphoric and neutral images from the \
        International Affective Picture System (IAPS; [28]). Each POFA and IAPS scene will be equalized (12.0 cd/m2) to match \
        mean luminance distribution. Task-related screens will be matched to mean luminance as well. Each trial will begin with \
        the appearance of a central fixation cross (FC) for 1500 ms, followed by an image pair. POFA pairs will be presented \
        for 3000 ms and IAPS pairs will be presented for 4500 ms. IAPS images will be presented for a longer duration because \
        they are typically more complex than the POFA faces. Following offset of the images, a small single or double asterisk \
        probe will appear in the location of one of the images and remain on the screen until the participant indicates whether \
        they detected one or two asterisks, with a maximum duration of 10,000 msec. Latency and accuracy of each response will \
        be recorded."

        #----exclude
        _exc = len(config['metadata']['subjects']['exclude'])
        _pct = (round(len(config['metadata']['subjects']['exclude'])/len(config['metadata']['subjects']['eyetracking']), 4)*100)
        m_['def']['exclude'] = "Participants with 'Dotloc' or 'Stimulus' Onset Error median above 3SD (<i>n</i> = %s, %.1f%%) \
        were excluded from analysis \
        (see <a class='anchor' href='../../methods.html#outliers'>methods</a>)."%(_exc, _pct)

        #----stimuli
        # iaps
        name = 'iaps'
        title = 'International Affective Picture System (IAPS)'
        ref = 'Lang, P., Bradley, M., Cuthbert, B. (2008). International affective picture system (IAPS): Affective \
        ratings of pictures and instruction manual. Technical Report A-8. University of Florida, Gainesville, FL.'
        m_['var'][name] = name
        m_['short'][name] = 'IAPS'
        m_['long'][name] = title
        m_['cite'][name] = ref
        m_['img'][name] = 'img/iaps.png'
        # pofa
        name = 'pofa'
        title = 'Pictures of Facial Affect (POFA)'
        ref = 'Ekman, P., Friesen, W., (1976). Pictures of facial affect. Consulting Psychologist Press, Palo Alto.'
        m_['var'][name] = name
        m_['short'][name] = 'POFA'
        m_['long'][name] = title
        m_['cite'][name] = ref
        m_['img'][name] = 'img/pofa.png'

        #----variables
        # os
        m_['var']['os'] = 'os'
        m_['short']['os'] = 'Operating System'
        m_['long']['os'] = 'Operating System (Microsoft OS, Mac OS)'
        m_['def']['os'] = 'Microsoft OS, Mac OS'
        #stimulus
        m_['var']['trialType'] = 'trialType'
        m_['short']['trialType'] = 'Stimulus'
        m_['long']['trialType'] = 'Stimulus (IAPS, POFA)'
        m_['def']['trialType'] = '<b>%s</b>: %s<br><b>%s</b>: %s'\
        %(m_['short']['iaps'], m_['long']['iaps'], m_['short']['pofa'], m_['long']['pofa'])
        # aoi
        m_['var']['aoi'] = 'aoi'
        m_['short']['aoi'] = 'Area of Interest'
        m_['long']['aoi'] = 'Area of Interest (sad, neutral images)'
        m_['def']['aoi'] = 'sad, neutral (categorical).'
        # participant
        m_['var']['participant'] = 'participant'
        m_['short']['participant'] = 'Participant'
        m_['long']['participant'] = 'Participant'
        m_['def']['participant'] = 'Participant number (categorical).'
        # trial
        m_['var']['TrialNum'] = 'TrialNum'
        m_['short']['TrialNum'] = 'Trial'
        m_['long']['TrialNum'] = 'Trial'
        m_['def']['TrialNum'] = ': 0-197 (categorical).'
        # cesd group
        m_['var']['cesd_group'] = 'cesd_group'
        m_['short']['cesd_group'] = 'CESD Score'
        m_['long']['cesd_group'] = "CESD Score ('Low' (&lt16) and 'High' (≥16))"
        m_['def']['cesd_group'] = ": Binary measure of CESD score ('Low' (&lt16) and 'High' (≥16))"
        # diff_stim
        m_['var']['diff_stim'] = 'diff_stim'
        m_['short']['diff_stim'] = 'Stimulus Onset Error'
        m_['long']['diff_stim'] = 'Stimulus Onset Error'
        m_['def']['diff_stim'] = ": The difference in time (msec) between predicted and true 'Stimulus' onset."
        # diff_dotloc
        m_['var']['diff_dotloc'] = 'diff_dotloc'
        m_['short']['diff_dotloc'] = 'Dotloc Onset Error'
        m_['long']['diff_dotloc'] = 'Dotloc Onset Error'
        m_['def']['diff_dotloc'] = ": The difference in time (msec) between predicted and true 'Dotloc' onset."
        # dotprobe bias score
        m_['var']['dp_bias'] = 'dp_bias'
        m_['short']['dp_bias'] = 'Dotprobe Bias Score'
        m_['long']['dp_bias'] = 'Dotprobe Bias Score'
        m_['def']['dp_bias'] = ': Trial level bias score, using the weighted method. From the R itrak package.'
        # dotprobe bias score
        m_['var']['gaze_bias'] = 'gaze_bias'
        m_['short']['gaze_bias'] = 'Gaze Bias Score'
        m_['long']['gaze_bias'] = 'Gaze Bias Score'
        m_['def']['gaze_bias'] = ': Trial level bias score, using the weighted method. From the R itrak package.'
        # dwell_time
        m_['var']['dwell_time'] = 'dwell_time'
        m_['short']['dwell_time'] = 'Dwell Time'
        m_['long']['dwell_time'] = 'Dwell Time'
        m_['def']['dwell_time'] = ": The duration (msec) during a trial gaze has been within an area of interest."

        #----modelling
        #lmer
        name = 'lmer'
        title = 'Bates, Mächler, Bolker, & Walker, 2015'
        doi = 'https://doi.org/10.18637/jss.v067.i01'
        ref = 'Bates, D., Mächler, M., Bolker, B., & Walker, S. (2015). Fitting linear mixed-effects models using lme4. 2015, 67(1), 48. \
        doi: %s'%(doi)
        m_['short'][name] = name
        m_['long'][name] = title
        m_['cite'][name] = ref
        m_['url'][name] = doi
        #glmer
        name = 'glmer'
        title = 'Bates, Mächler, Bolker, & Walker, 2015'
        doi = 'https://doi.org/10.18637/jss.v067.i01'
        ref = 'Bates, D., Mächler, M., Bolker, B., & Walker, S. (2015). Fitting linear mixed-effects models using lme4. 2015, 67(1), 48. \
        doi: %s'%(doi)
        m_['short'][name] = name
        m_['long'][name] = title
        m_['cite'][name] = ref
        m_['url'][name] = doi
        #anova
        name = 'anova'
        title = 'Chambers, Hastie, 1992'
        ref = 'Chambers, J., Hastie, T. (1992). Statistical Models in S. Wadsworth & Brooks/Cole.'
        m_['short'][name] = name
        m_['long'][name] = title
        m_['cite'][name] = ref

    	#----assumptions
        # lmer
    	## residual normal distribution
        m_['long']['rnd'] = "Residuals are normally distributed"
        m_['def']['rnd'] = ": Residuals are drawn from a normally distributed population. Can be checked with a QQ plot. \
        Statistical tests, such as Anderson-Darling and Kolmogorov–Smirnov are also possible."
    	## independent observations
        m_['long']['rio'] = "Residuals are independent"
        m_['def']['rio'] = ": Residuals have been drawn independently of each other. This can be checked by plotting residuals \
        against covariates - especially time-varying or spatial covariates."

        # anova
        ## normal distribution
        m_['long']['nd'] = "Samples are normally distributed"
        m_['def']['nd'] = ": Samples are drawn from a normally distributed population."
    	## Homogeneity of variance
        m_['long']['hv'] = "Homogeneity of variance"
        m_['def']['hv'] = ": Amount of variance for a IV is constant across the sample."
    	## independent observations
        m_['long']['io'] = "Independent observations"
        m_['def']['io'] = ": Samples have been drawn independently of each other."

        # Estimated Marginal Means
        url = 'https://cran.r-project.org/web/packages/emmeans/vignettes/comparisons.html'
        m_['def']['emm'] = "The 'Estimated Marginal Means' table provides a list pairwise comparisons for each level and factor \
        in our model (<a class='anchor' href='%s'>emmeans package, Version 1.3.2</a>)."%(url)
    	# residuals vs fitted
        m_['def']['rvf'] = "The Residuals vs Fitted plot is used to identify residual distribution. \
        If residuals are following a visible trend on the graph, then the homogeneity assumption was violated."
    	# qq
        m_['def']['qq'] = "The Normal Q-Q plot compares the standardized residuals against the theoretical quantiles from \
        a standard normal distribution. If the model residuals are normally distributed, then the points \
        on this graph will be plotted in a generally straight line."

        #----calibration, validation
        m_['def']['calibration'] = "Symbols: square = mouse click, circle = calibration point, cross = gaze coordinates.<br>\
        To use the interactive plot, drag the mouse along the calibration point chart. Each pair of dots repersents gaze coordinates \
        recorded during the presentation of one of the calibration points."
        m_['def']['validation'] = "Symbols: circle = calibration point, cross = gaze coordinates.<br>\
        To use the interactive plot, drag the mouse along the calibration point chart. Each pair of dots repersents gaze coordinates \
        recorded during the presentation of one of the calibration points."

        #----data processing: classify, filter (processing.py)
        # hmm
        name = 'hmm'
        title = 'Pekkanen & Lappi, 2017'
        doi = 'doi: 10.1038/s41598-017-17983-x'
        ref = 'Pekkanen, J., & Lappi, O. (2017). A new and general approach to signal denoising and eye movement classification \
        based on segmented linear regression. Scientific Reports, 7(1).'
        m_['short'][name] = name
        m_['long'][name] = title
        m_['cite'][name] = ref
        m_['url'][name] = doi
        # how preprocessing was planned
        name = 'preprocessing'
        title = 'Larsson, Nystrom, & Stridh, 2013'
        doi = 'doi: 10.1109/tbme.2013.2258918'
        ref = 'Larsson, L., Nystrom, M., & Stridh, M. (2013). Detection of Saccades and Postsaccadic Oscillations in the Presence \
        of Smooth Pursuit. IEEE Transactions On Biomedical Engineering, 60(9), 2484-2493.'
        m_['short'][name] = name
        m_['long'][name] = title
        m_['cite'][name] = ref
        m_['url'][name] = doi
        # example
        m_['def']['filter'] = "As conventional in eye tracking research, eye movement data were filtered according to \
        predetermined cut-offs (e.g. Juhasz and Rayner, 2006; Kliegl, Grabner, Rolfs, and Engbert, 2004; Rayner, Reichle, Stroud, \
        Williams, and Pollatsek, 2006; Rayner et al., 2011; Schattka, Radach, and Huber, 2010). Fixations smaller than 80 ms \
        and adjacent to a larger neighbouring fixation (within 0.3° of a visual angle) were merged. Fixations shorter than \
        80 ms that were not adjacent to a larger neighbouring fixation and fixations longer than 1200 ms were excluded \
        (Juhasz et al., 2006). Trials with gross track loss were also excluded, leading to an elimination of 4.47% of the \
        data overall (3.68% for NHI and 4.80% for PWA). Trials with blinks were not excluded since a comparison of trials \
        with blinks to trials without blinks showed no significant differences in gaze durations in a previous experiment (Huck, 2016)."

        #----software
        name = 'R'
        title = 'R Core Team, 2013'
        url = 'http://www.R-project.org/'
        ref = 'Core Team, R., (2013). R: A language and environment for statistical computing. R Foundation for Statistical Computing, Vienna,\
        Austria. (Version 3.2.1) [Software]. Available from URL %s.'%(url)
        m_['short'][name] = name
        m_['long'][name] = title
        m_['cite'][name] = ref
        m_['url'][name] = url

        #----remove whitespaces
        for key, value in m_['def'].items():
            m_['def'][key] = re.sub(r'\s+', ' ', value).strip()
        for key, value in m_['cite'].items():
            m_['cite'][key] = re.sub(r'\s+', ' ', value).strip()

        #create metadata variable
        config['metadata']['short'] = m_['short']
        config['metadata']['long'] = m_['long']
        config['metadata']['def'] = m_['def']
        config['metadata']['cite'] = m_['cite']
        config['metadata']['url'] = m_['url']
        config['metadata']['var'] = m_['var']
        config['metadata']['img'] = m_['img']

        return config




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| `@purpose`: mdl, an extensive library for the exploration, visualization, and analysis of eyetracking data. \
This library was created at the Institute for Mental Health Research, at the University of Texas at Austin by Semeon Risom.  
| `@date`: Created on Sat May 1 15:12:38 2019  
| `@author`: Semeon Risom  
| `@email`: semeon.risom@gmail.com  
| `@url`: https://semeon.io/d/mdl
"""
# allowed imports
__all__ = ['download','eyetracking','plot','r33','settings']

# core
from pdb import set_trace as breakpoint
import os
import sys
import certifi

# relative paths
sys.path.append(os.path.dirname(os.path.realpath(__file__)))

# set as namespace package
# see https://stackoverflow.com/questions/41621131/python-namespace-packages-in-python3
try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
    del pkg_resources
except ImportError:
    import pkgutil
    __path__ = pkgutil.extend_path(__path__, __name__)
    del pkgutil

# imports
from . import download
from . import eyetracking
from . import plot
from . import r33
from . import settings

# del os, sys

# version
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
@`purpose`: Create regions of interest to export into Eyelink DataViewer or statistical resources such as R and python.  
@`date`: Created on Sat May 1 15:12:38 2019  
@`author`: Semeon Risom  
@`email`: semeon.risom@gmail.com  
@`url`: https://semeon.io/d/mdl
"""
#%% [markdown]
# #### Read ROI from photoshop PSD files
#%% [markdown]
# ##### import mdl package
#%%
import mdl
#%% [markdown]
# ##### set path
#%%
image_path = './raw/'
output_path = './output/'
metadata_source = 'metadata.xlsx'
#%% [markdown]
# ##### initiate
#%%
roi = mdl.eyetracking.ROI(isMultiprocessing=False, isDebug=True, isLibrary=False, isDemo=True,
	image_path=image_path, output_path=output_path, metadata_source=metadata_source, 
	scale=1, screensize=[1920,1080], center=[(1920*.5),(1080*.5)], shape='straight', 
	roi_format='both', uuid=['image','roi','position'], newcolumn={'position': 'center'})
#%% [markdown]
# ##### start
#%%
df, error = roi.process()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| @purpose: Hub for running processing and analyzing raw data.   
| @date: Created on Sat May 1 15:12:38 2019   
| @author: Semeon Risom   
| @email: semeon.risom@gmail.com   
| @url: https://semeon.io/d/R33-analysis   
"""

# available classes and functions
__all__ = ['Processing']

# required external libraries
__required__ = ['distutils','importlib','nslr']

# core
from pdb import set_trace as breakpoint
from pathlib import Path
import numpy as np
import pandas as pd
import math as m
import sys
import os
import glob

## logging
import logging
import datetime

# local libraries
if __name__ == '__main__':
	from .. import Settings
	from .. import Metadata

class Processing():
    """Hub for running processing and analyzing raw data."""
    def __init__(self, config, isLibrary=True):
        """
		Initiate the mdl.r33.Processing module.

        Parameters
        ----------
        config : :class:`dict`
            Configuration data. i.e. trial number, location.
        isLibrary : :obj:`bool`
            Check if required libraries are available. Default `False`.
        """

        #check libraries
        if isLibrary:
            settings.library(__required__)

        #set current subject (use for iterations)
        self.current_subject = ''
        self.cgxy = ''
        self.log = ''
        self.thisCore = 0
        self.config = config
        self.filters = [['SavitzkyGolay', 'sg']]
		
    def getEstimatedMonitor(self, diagonal, window):
        """calculate estimate monitor size (w,h;cm) using estimated diagonal monitor (hypotenuse; cm).

        Attributes
        ----------
        df_raw : :class:`pandas.DataFrame`
            Pandas dataframe of subjects.
        """
        ratio = window[0]/window[1]
        monitor_x = float((ratio * (m.sqrt((m.pow(diagonal,2)) / (m.pow(ratio,2) + 1)))) * 2.54)
        monitor_y = float((m.sqrt((m.pow(diagonal,2)) / (m.pow(ratio,2) + 1))) * 2.54)
        monitor = [monitor_x, monitor_y]
        
        return monitor

    def preprocess(self, df, window):
        """Initial data cleaning.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Pandas dataframe of raw data.
        window : :class:`tuple`
            horizontal, vertical resolution

        Attributes
        ----------
        m_delta : :obj:`int`
            Maxinum one-sample change in velocity

        Notes
        ----------
        remove_missing:
            Remove samples with null values.
        remove_bounds:
            Remove samples outside of window bounds (1920,1080).
        remove_spikes:
            remove one-sample spikes if x and y-axis delta is greater than 5.
        """
        #sort
        df = df.sort_values(['TrialNum','sampleNum'],ascending=[True, True])

        #1) remove fixation cross samples #!!!
        #print(self.console['green'] + '///////////Preprocessing: Drop Fixation Cross samples' + self.console['ENDC'])
        #df = df[df['event'] != 'Fixation'].reset_index(drop=True)

        #2) get RT from timestamp
        d_loc = df[df['event'] == "DotLoc"]
        #check if participant responded
        if d_loc.shape[0] >=1:
            df["RT"] = d_loc['timestamp'][d_loc.index[-1]] - d_loc['timestamp'][d_loc.index[0]]
        else:
            self.log.warning('event: preprocess(self, df, window), core: %s, subject: %s, trial: %s, dotloc rows: %s'
                             %(self.thisCore, self.config['subject'], self.config['trial'], d_loc.shape[0]))

        #3) set event markers #!!!
        df['marker'] = "."

        #fixation event
        ##set marker
        #df['marker'][df.index[0]] = "Fixation Onset"
        df.loc[df.index[0], 'marker'] = "Fixation Onset"

        #stim event
        d_stim = df[df['event'] == "Stim"]
        ##set marker
        df.loc[d_stim.index[0], 'marker'] = "Stimulus Onset"

        #dotloc event #!!!
        ##set marker
        ##check if any samples exist within dotloc onset
        if d_loc.shape[0] >=1:
            df.loc[d_loc.index[0], 'marker'] = "Dotloc Onset"
        else:
            self.log.warning('event: preprocess(self, df, window), core: %s, subject: %s, trial: %s, dotloc rows: %s'
                             %(self.thisCore, self.config['subject'], self.config['trial'], d_loc.shape[0]))
            
        #end trial
        #df['marker'][df.index[-1]] = "End Trial"
        df.loc[df.index[-1], 'marker'] = "End Trial"

        #4) remove missing data #!!!
        ## convert '.' to null
        if self.config['remove_missing']:
            print(self.console['green'] + 'Preprocessing: Remove null samples' + self.console['ENDC'])
            ## drop null
            df = df[(df["x"].notnull()) & (df["y"].notnull())].reset_index(drop=True)
        else:
            #mark as bad sample
            df['bad'] = df.apply(lambda x: True if ((x['x'] == '.') or (x['y'] == '.')) else False, axis=1)

        #5) remove samples outside of window bounds (i.e. 0>x>1920, 0>y>1080) #!!!
        if self.config['remove_bounds']:
            print(self.console['green']+'Preprocessing: Remove samples outside of window bounds (i.e. 0 > x > 1920, 0 > y > 1080)'\
                  +self.console['ENDC'])
            df = df[(df['x'] <= window[0]) & (df['y'] <= window[1]) &
                    (df['x'] >= 0) & (df['y'] >= 0)].reset_index(drop=True)
        else:
            #mark as bad sample
            df['bad'] = df.apply(lambda x: True if ((x['x'] <= 0) or (x['y'] <= 0) or
                                                    (x['x'] > window[0]) or (x['y'] > window[1])
                                                    ) else x['bad'], axis=1)

        #6) remove one-sample spikes #!!!
        if self.config['remove_spikes']:
            print(self.console['green'] + 'Preprocessing: Remove one-sample spikes' + self.console['ENDC'])
            max_delta = self.config['spike_delta']
            #x
            delta = df['x'].diff().shift(-1).abs()
            delta.iloc[-1] = delta.iloc[-2]
            df = df[(delta < max_delta)]
            #y
            delta = df['y'].diff().shift(-1).abs()
            delta.iloc[-1] = delta.iloc[-2]
            df = df[(delta < max_delta)]
        # else mark spikes as bad
        else: 
            max_delta = self.config['spike_delta']
            #x
            delta = df['x'].diff().shift(-1).abs()
            delta.iloc[-1] = delta.iloc[-2]
            df_spike = df[(delta < max_delta)]
            #y
            delta = df['y'].diff().shift(-1).abs()
            delta.iloc[-1] = delta.iloc[-2]
            df_spike = df_spike.loc[(delta < max_delta)]

            #mark samples as bad
            spike = df_spike.index
            df.loc[spike,'bad'] = True

        #set x,y to NaN
        #df["x"] = df["x"].replace(".", np.NaN)
        #df["y"] = df["y"].replace(".", np.NaN)

        return df

    #getting data for analysis

    def getData(self, path=None):
        """preparing data for use in analysis

        Parameters
        ----------
        path : :obj:`str`
            The directory path of the subject data

        Attributes
        ----------
        path : :obj:`str`
            Specific directory path used.

        Returns
        -------
        df : :class:`pandas.DataFrame`
            Pandas dataframe of raw data.
        _path : :obj:`list`
            list of files used for analysis.

        Notes
        -----
        You can either get data from all subjects within a directory, or from a specific subject (subject_session).

        Examples
        --------
        >>> #if using path:
        >>> df_raw = getData(path=self.config['path'])

        >>> #if getting data for single subject:
        >>> df_raw = getData(path=self.config['path'],subject_session=['1099','1', '0'])

        """
        #if single subject get path from config
        if (self.config['single_subject']):
            path = '%s/%s_%sabc.csv'%(self.config['path'],self.config['subject'],self.config['session'])
        
        #read csv as dataframe
        df = pd.read_csv(path, float_precision='high')

        ##rename
        if self.config['source'] != 'eyelink':
            df = df.rename(columns={"trialNumTask": "TrialNum", "sample_time": "timestamp",
                                    "isWindowSuccess": "is_calibrated",
                                    "LEmotion": "left_mood","REmotion": "right_mood"})

        #sort    
        df = df.sort_values(['TrialNum','timestamp'],ascending=[True, True]).reset_index(drop=True)

        #set as eyetracking or behavioral
        if self.config['source'] != 'eyelink':
            ##create column for type (eyetracking, behavioral)
            df['type'] = np.where((df['isWebcamUsed']==True), 'eyetracking', 'behavioral')

        return df, path

    def filter_data(self, df, filter_type, config, **kwargs):
        """
        Butterworth: Design an Nth-order digital or analog Butterworth filter and return
        the filter coefficients.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Pandas dataframe of raw data.
        filter_type : :obj:`str`, optional
			Type of filter.
        config : :class:`dict`
            Configuration data. i.e. trial number, location.

        Attributes
        ----------
        filter_type : :obj:`str`
            Filter type: 'butterworth'

        """
        from scipy.ndimage.filters import gaussian_filter1d
        from scipy.signal import butter,filtfilt,medfilt,savgol_filter

        g_t= df['timestamp']
        g_x = df['x']
        g_y= df['y']

        """filters"""
        #http://scipy-cookbook.readthedocs.io/items/ButterworthBandpass.html
        #Butterworth filter
        #note: butterworth cant have NaN data
        if filter_type=='butterworth':
            wn=self.config['f_b']['Wn']
            order=self.config['f_b']['N']
            b, a = butter(order, wn, 'lowpass', output='ba')
            #b, a = butter(order, wn, 'low')
            #converting pandas series to numpy ndarray
            bg_t=df['timestamp']
            bg_x=df['x']
            bg_y=df['y']
            f_x = filtfilt(b, a, bg_x)
            f_y = filtfilt(b, a, bg_y)
            #print(f_x)
            fxy_df = pd.DataFrame({'x':f_x, 'y':f_y, 'timestamp':bg_t})
            #breakpoint() #TODO!

        #gaussian filter
        elif filter_type=='gauss':
            sigma=self.config['f_g']['sigma']
            print('sigma: %s'%(sigma))
            f_x = gaussian_filter1d(g_x,sigma)
            f_y = gaussian_filter1d(g_y,sigma)
            fxy_df = pd.DataFrame({'x':f_x, 'y':f_y, 'timestamp':g_t})

        #median filter
        elif filter_type=='median':
            size=self.config['f_m']['size']
            print('size: %s'%(size))
            f_x=medfilt(g_x,size)
            f_y=medfilt(g_y,size)
            fxy_df = pd.DataFrame({'x':f_x, 'y':f_y, 'timestamp':g_t})

        #http://scipy-cookbook.readthedocs.io/items/SavitzkyGolay.html
        #savitzky golay filter
        elif filter_type=='SavitzkyGolay':
            window=self.config['f_sg']['window']
            order=self.config['f_sg']['order']
            print(self.console['orange'] + 'window: %s'%(window) + self.console['ENDC'])
            print(self.console['orange'] + 'order: %s'%(order) + self.console['ENDC'])
            #breakpoint() #TODO!
            #f_x=analysis.savitzky_golay(g_x,window_size=window, order=order)
            #f_y=analysis.savitzky_golay(g_y,window_size=window, order=order)
            try:
                f_x=savgol_filter(g_x, window_length=window, polyorder=order)
                f_y=savgol_filter(g_y, window_length=window, polyorder=order)
                fxy_df = pd.DataFrame({'x':f_x, 'y':f_y, 'timestamp':g_t})
            except:
                fxy_df = None

        #mean filter
        elif filter_type=='moving average':
            weights=self.config['f_a']['weights']
            print('weights: %s'%(weights))
            weights=weights/np.sum(weights)
            print('weights: %s'%(weights))
            f_x=np.convolve(g_x, weights,'same')
            f_y=np.convolve(g_y, weights,'same')
            fxy_df = pd.DataFrame({'x':f_x, 'y':f_y, 'timestamp':g_t})

        else:
            raise ValueError('Unknown Filter Type: %s. Must be one of %s'%(filter_type,str(
                    ['sg','butter','gauss','median','moving average'])))

        return fxy_df

    def classify(self, config, df, ctype='ivt', filter_type=None,
                 v_th=None, dr_th=None, di_th=None,
                 missing=None, maxdist=None, mindur=None):
        """
        I-DT algorithm takes into account the distribution or spatial proximity of
        eye position points in the eye-movement trace.

        In the I-VT model, the velocity value is computed for every eye position
        sample. The velocity value is then compared to the threshold. If the
        sampled velocity is less than the threshold, the corresponding eye-position
        sample is marked as part of a fixation, otherwise it is marked as a part of
        a saccade.

        The simple model detects fixations, defined as consecutive samples with an
        inter-sample distance of less than a set amount of pixels (disregarding missing data)

        Parameters
        ----------
        config : :class:`dict`
            Configuration data. i.e. trial number, location.
        df : :class:`pandas.DataFrame`
            Pandas dataframe of classified data.
        ctype : :obj:`str`
            Classification type: 'ivt'
        filter_type : [type], optional
            Filter type: 'butter'
        ctype : :obj:`int`, optional
            velocity threshold (ivt), dispersion threshold (idt; used by
            SR-Research and Tobii), or simple
        v_th : :obj:`str`
            Velocity threshold in pix/sec (ivt)
        dr_th : :obj:`str`
            Fixation duration threshold in pix/msec (idt)
        di_th : :obj:`str`
            Dispersion threshold in pixels (idt)
        missing : :obj:`str`
            value to be used for missing data (simple)
        maxdist : :obj:`str`
            maximal inter sample distance in pixels (simple)
        mindur : :obj:`str`
            minimal duration of a fixation in milliseconds; detected fixation
            cadidates will be disregarded if they are below this duration  (simple)

        Raises
        ------
        ValueError
            Unknown classification type.

        Returns
        -------
        df : :class:`pandas.DataFrame`
            Pandas dataframe of classified data.
        """
        from . import Classify

        if ctype == 'ivt':
            cnfg = self.config
            df = Classify.ivt(df, v_th, config=cnfg)

        elif ctype == 'idt':
            df = Classify.idt(df, di_th, dr_th)

        elif ctype == 'hmm':
            cnfg = self.config
            df_, cxy_df = Classify.hmm(data=df, config=cnfg, filter_type=filter_type)
            df = [df_, cxy_df]

        elif ctype == 'simple':
            df = Classify.simple(df, missing, maxdist, mindur)

        else:
            raise ValueError('Unknown classification type: %s. Must be one of %s' % (
                ctype, str(['ivt', 'idt'])))

        return df

    def roi(self, filters=[['SavitzkyGolay','sg']], flt=None, df=None, manual=False, monitorSize=None):
        """Check if fixation is within bounds.

        Attributes
        ----------
        manual : :class:`str`
            Whether or not processing.roi() is access manually.
        monitorSize : :class:`list`
            Monitor size.
        filters : :class:`list`
            Filter parameters.
        df : :class:`pandas.DataFrame`
            Pandas dataframe of classified data.

        Returns
        -------
        df : :class:`pandas.DataFrame`
            Pandas dataframe of classified data.
        """
        #timestamp
        t0 = datetime.datetime.now()
        function = self.debug(message='t', source="timestamp")
        
        for idx, itm in enumerate(filters):
            #filter type
            flt = itm[1]
            print(self.console['orange'] + 'bounds: %s'%(itm) + self.console['ENDC'])
            
            #if not running model.roi manually (posthoc)
            if (manual == False):
                #real bounding box
                monitorSize=self.config['resolution.px']
            else:
              monitorSize = [int(i) for i in (df['monitorSize.px'].values[0]).split('x')]  

            #if x-resolution is smaller than 1400
            if monitorSize[0] < 1400:
                scaleImage = self.config['scaleImage']
                #scale container
                c = [1366*scaleImage,768*scaleImage]
                #scale stimulus
                image_x = 600*scaleImage #image size
                image_y = 600*scaleImage #image size
            #dont scale stimulus but create bounds larger than stimulus
            else:
                #scale roi #percentage of screen devoted to stim if window<=1400
                #scaleROI = (1 + (600/1400))
                #container
                c = [1366,768]
                #scale stimulus
                image_x = 600 #image size
                image_y = 600 #image size

            #centers
            cx_center = c[0]/2 #container x-center
            image_y_c = image_y/2 #roi bound y-center
            #bound_y_c = bound_y/2 #image y-center
            resx_c = monitorSize[0]/2 #resolution x-center
            resy_c = monitorSize[1]/2 #resolution y-center

            #------------------------------------------creating stim bounds
            #----------left stim bound
            lsbx1, lsbx2 = (resx_c-cx_center, resx_c-cx_center+image_x)
            lsby1, lsby2 = (resy_c+image_y_c, resy_c-image_y_c)
            #----------right stim bound
            rsbx1, rsbx2 = (resx_c+cx_center-image_x, resx_c+cx_center)
            rsby1, rsby2 = (resy_c+image_y_c, resy_c-image_y_c)
            #add to list
            stim_bounds = [dict(bID='l', bx1=lsbx1,by1=lsby1,bx2=lsbx2,by2=lsby2),
                     dict(bID='r', bx1=rsbx1,by1=rsby1,bx2=rsbx2,by2=rsby2)]
            
            #if not running processing.roi() manually (posthoc)
            if (manual == False):
                self.config['stim_bounds'] = stim_bounds

            #------------------------------------------creating roi bounds
            #----------left roi bound
            ##---region
            lbx1, lbx2 = (0, lsbx2)
            lby1, lby2 = (monitorSize[1], 0)
            ##---stim
            #lbx1, lbx2 = (lsbx1, lsbx2)
            #lby1, lby2 = (lsby2, lsby1)
            #----------right roi bound
            ##---region
            rbx1, rbx2 = (rsbx1, monitorSize[0])
            rby1, rby2 = (monitorSize[1], 0)
            ##---stim
            #rbx1, rbx2 = (rsbx1, rsbx2)
            #rby1, rby2 = (rsby2, rsby1)
            #add to list
            roi_bounds = [dict(bID='l', bx1=lbx1,by1=lby1,bx2=lbx2,by2=lby2),
                     dict(bID='r', bx1=rbx1,by1=rby1,bx2=rbx2,by2=rby2)]
            
            #if not running processing.roi() manually (posthoc)
            if (manual == False):
                self.config['roi_bounds'] = roi_bounds

            #if using eyelink data and eyelink classification
            if ((self.config['source'] == 'eyelink') and (self.config['classify_eyelink_data'] == False)):
                pts = np.array(df[['x', 'y']])
            #else use xy coordintes used to classify
            else:
                pts = np.array(df[['%s_x'%(flt), '%s_y'%(flt)]])

            #---------------------------------------------------------------------------------------------left bounds
            itm0 = roi_bounds[0] #x,y coordinates
            L_Bll = np.array([itm0['bx1'], itm0['by2']]) # lower-left
            L_Bur = np.array([itm0['bx2'], itm0['by1']]) # upper-right
            #bool of coordinates within and outside of bounds
            left_bound = np.all(np.logical_and(L_Bll <= pts, pts <= L_Bur), axis=1)

            #-------------------------------------------------------------------------------------------right bounds
            itm1 = roi_bounds[1] #x,y coordinates
            R_Bll = np.array([itm1['bx1'], itm1['by2']]) # lower-left
            R_Bur = np.array([itm1['bx2'], itm1['by1']]) # upper-right
            #bool of coordinates within and outside of bounds
            right_bound = np.all(np.logical_and(R_Bll <= pts, pts <= R_Bur), axis=1)

            #if not running processing.roi() manually (posthoc)
            if (manual == False):
                #-----------------------------------------------------------------get roi
                df = pd.concat([df, pd.DataFrame(np.vstack((left_bound, right_bound)).T,\
                                                 columns=['left_bound','right_bound'])], axis=1, sort=False)
                #if eyelink data
                if ((self.config['source'] == 'eyelink') and (self.config['classify_eyelink_data'] == False)):
                    #all fixation events
                    df['%s_fix_all'%(flt)] = df.apply(lambda x: True if (isinstance(x['fixation'], int))\
                                                      else False, axis=1)
                    #-------roi
                    #gaze and fixations within roi
                    df['%s_all_bounds'%(flt)] = df.apply(lambda x: 1 if (x['left_bound'] == True)\
                                                         else (2 if (x['right_bound'] == True) else False), axis=1)
                    ##check if sample is within bounds and part of a fixation event
                    df['%s_fix_bounds'%(flt)] = df.apply(lambda x: x['%s_all_bounds'%(flt)]\
                                                      if (isinstance(x['fixation'], int)) else False, axis=1)
                #webgazer data
                else:
                    #If enough samples
                    if (self.config['too_few_samples'] == False):
                        #-----------------------------------------------------------------all fixation events
                        df['%s_fix_all'%(flt)] = df.apply(lambda x: True if (x['%s_class'%(flt)]==1) else False, axis=1)
    
                        #fixation index - count of fixation events per trial (i.e. fixation: 1, 2, 3, 4)
                        ##drop non fixations and get index
                        p_df = df.drop(df[df['%s_fix_all'%(flt)] == False].index)
                        #-----------------------------------------------------------------fixation index
                        ##add index to dataframe #subset dataframe to single column
                        p_df['%s_fix_index'%(flt)] = range(len(p_df))
                        p_df = p_df[['%s_fix_index'%(flt)]]
                        ##add column to original dataframe
                        df = df.join(p_df)
                        #replace np.nan with "." (for R)
                        # df['%s_fix_index'%(flt)].where((df['%s_fix_index'%(flt)].notnull()), ".", inplace=True)
                        #------------------------------------fixation counter
                        #enumerate
                        df['enum'] = ((df['%s_fix_all'%(flt)] != df['%s_fix_all'%(flt)].shift(1)).cumsum()-1)
                        #reset non-fixations as None
                        df['fix_num'] = df.apply(lambda x: float(x['enum']) if (x['%s_fix_index'%(flt)] != '.') else None, axis=1)
                        #factorize and reset #finished
                        df['fix_num'] = pd.factorize(df['fix_num'])[0]
                        df['fix_num'] = df.apply(lambda x: float(x['fix_num']) if (x['%s_fix_index'%(flt)] != '.') else None, axis=1)
                        #------------------------------------roi
                        ##gaze and fixations within roi
                        df['%s_all_bounds'%(flt)] = df.apply(lambda x: 1 if (x['left_bound'] == True)\
                                                             else (2 if (x['right_bound'] == True) else "."), axis=1)

                        ##only fixation within roi
                        df['%s_fix_bounds'%(flt)] = df.apply(lambda x: x['%s_all_bounds'%(flt)]\
                                                          if (x['%s_class'%(flt)]!=None) else ".", axis=1)
                
                        ##only fixation within roi
                        df['%s_fix_bounds_old'%(flt)] = df.apply(lambda x: x['%s_all_bounds'%(flt)]\
                                                          if (x['%s_class'%(flt)]==1) else ".", axis=1)
                        
                        #------------------------------------total samples (left and right dwell) counter
                        df['sample_total'] = df.shape[0]
                    
                        #----calculate dwell time
                        df['dwell'] = df.apply(lambda x: True if (x['left_bound']==True or x['right_bound']==True) else False, axis=1)

                    #not enough samples
                    else:
                        not_enough_samples = 'subject: %s, trial: %s; not enough samples'%(self.config['subject'], self.config['trial'])
                        self.debug(not_enough_samples)
                        print(self.console['red'] + not_enough_samples + self.console['ENDC'])
                        #add blank for fixation
                        df['%s_fix_index'%(flt)] = "."
                        df['%s_fix_all'%(flt)] = "."
                        df['fix_num'] = "."
                        df['%s_all_bounds'%(flt)] = "."
                        df['%s_fix_bounds'%(flt)] = "."
                        df['%s_fix_bounds_old'%(flt)] = "."
                        df['dwell'] = "."
                        df['sample_total'] = df.shape[0]
            #----------------------------------------------------------------------------------------test confirmation
            test = False
            if test and (self.config['subject'] != self.current_subject):
                self.current_subject = self.config['subject']
                import matplotlib
                matplotlib.use('Agg')
                import matplotlib.pyplot as plt
                import matplotlib.patches as patches
                print(self.console['green'] + 'Preprocessing: Test to ensure ROI' + self.console['ENDC'])

                #plot
                #draw cross ###af2f2c #draw square ##2222b2 #draw triangle ##198D40
                plt.figure(4, figsize=(9.6,5.4))
                test = plt.gca()

                #roi bounding boxes #lbx1, lbx2, lby1, lby2 #rbx1, rbx2, rby1, rby2
                test.add_patch(patches.Rectangle((lbx1,lby2),width=lbx2,height=lby1,linewidth=1,facecolor='#396cbd33'))
                test.add_patch(patches.Rectangle((rbx1,rby2),width=rbx2,height=rby1,linewidth=1,facecolor='#396cbd33'))

                #gaze coordinates
                test.plot(df['%s_x'%(flt)], df['%s_y'%(flt)], marker='+', markerfacecolor="#ffffff00",
                          markersize=6, markeredgecolor='#af2f2c', linewidth=0)

                #all fixations
                p_df = df.drop(df[df['%s_fix_all'%(flt)] == False].index)
                test.plot(p_df['%s_x'%(flt)], p_df['%s_y'%(flt)], marker='s', markerfacecolor="#ffffff00",
                          markersize=6, markeredgecolor='#2222b2', linewidth=0)

                #only fixations within roi
                p_df = df[(df['%s_fix_all'%(flt)] == True) & (df['%s_fix_bounds'%(flt)] != ".")]
                test.plot(p_df['%s_x'%(flt)], p_df['%s_y'%(flt)], marker='^', markerfacecolor="#ffffff00",
                          markersize=6, markeredgecolor='#198D40', linewidth=0)

                #set limits
                test.set_xlim([0,monitorSize[0]])
                test.set_ylim([0,monitorSize[1]])

                #position
                ###########left_image x0, left_image x1, center, right_image x0, right_image x1
                x_ticks = [roi_bounds[0]['bx1'], roi_bounds[0]['bx2'], resx_c, roi_bounds[1]['bx1'], roi_bounds[1]['bx2']]
                ##########image y0, center, image y1
                y_ticks = [roi_bounds[0]['by1'], resy_c, roi_bounds[0]['by2']]
                #set ticks
                test.set_xticks(x_ticks)
                test.set_yticks(y_ticks)
                x_ticks.sort()
                y_ticks.sort()
                test.grid(linestyle='--')
                plt.gca().invert_yaxis()
                #plt.show()

                #save
                dpi = self.config['fig_dpi']
                emotion = self.config['emotion']
                resolution = self.config['resolution.px']
                img_title = 'gaze coordinates [subject:%s, session:%s, trial:%s, isCongruent:%s, left:%s, right:%s (%s, %s)]'\
                                                %(self.config['subject'],self.config['session'],self.config['trial'],\
                                                self.config['isCongruent'],emotion[0],emotion[1],\
                                                resolution[0],resolution[1])
                plt.suptitle(img_title, fontsize=12)
                save_folder = os.path.abspath(self.config['path']['output']+ '/test/img/%s_%s_%s-roiPlot.png'\
                                              %(self.config['subject'],self.config['session'],self.config['trial']))
                plt.savefig(save_folder, dpi=dpi,transparent=False)
                plt.close()
            
            #--------finished
            #timestamp
            print(self.console['blue']+'%s finished in %s msec'%(function,
                  ((datetime.datetime.now()-t0).total_seconds()*1000))+self.console['ENDC'])   
            #if not running processing.roi() manually (posthoc)         
            if (manual == False):
                return df
            else:
                return stim_bounds, roi_bounds

    def process(self, window, filters, gxy_df, trial, _classify=True, ctype='simple', _param='', log=False,
                 draw_plot=False,  draw_heatmap=False, draw_gazeplot=False, draw_fixplot=False,
                 v_th=20,dr_th=200,di_th=20,
                _missing=0.0, _maxdist=25, _mindur=50):
        """
        Plotting and preparing data for classification. Combined plot of each filter.

        Parameters
        ----------
        window : :obj:`list`
            horizontal, vertical resolution
        filters : :obj:`list`
            List of filters along with short-hand names.
        gxy_df : :class:`pandas.DataFrame`
            Pandas dataframe of raw data. Unfiltered raw data.
        trial : :obj:`str`
            Trial number.
        _classify : :obj:`bool`
            parameter to include classification
        ctype : :obj:`str`
            classification type. simple, idt, ivt
        _param : :obj:`str`
            [description] (the default is '', which [default_description])
        log : :obj:`bool`
            [description] (the default is False, which [default_description])
        draw_plot : bool, optional
            [description] (the default is False, which [default_description])
        draw_heatmap : bool, optional
            [description] (the default is False, which [default_description])
        draw_gazeplot : bool, optional
            [description] (the default is False, which [default_description])
        draw_fixplot : bool, optional
            [description] (the default is False, which [default_description])
        v_th : :obj:`str`
            Velocity threshold in px/sec (ivt)
        dr_th : :obj:`str`
            Fixation duration threshold in px/msec (idt)
        di_th : :obj:`str`
            Dispersion threshold in px (idt)
        _missing : :obj:`bool`
            value to be used for missing data (simple)
        _maxdist : :obj:`str`
            maximal inter sample distance in pixels (simple)
        _mindur : :obj:`str`
            minimal duration of a fixation in milliseconds; detected fixation
            cadidates will be disregarded if they are below this duration  (simple)
            (default = 100)

        Attributes
        ----------
        _fxy_df : :class:`pandas.DataFrame`
            Pandas dataframe of raw data. Filtered data. Subset of _fgxy_df.

        Returns
        -------
        _fgxy_df : :class:`pandas.DataFrame`
            Pandas dataframe of filtered data.
        c_xy : :class:`pandas.DataFrame`
                Pandas dataframe of classified data.
        """

        #%matplotlib inline
        #append filtered and raw data to list being prepared for export
        l_gf = []
        l_cxy = []
        #append raw data
        l_gf.append(gxy_df)
        
        for idx, itm in enumerate(filters):
            print(self.console['orange'] + 'filter: %s' % (itm[0]) + self.console['ENDC'])
            #-------------------------------------------------------------------------parameters
            c_xy = []  # classify data

            #if using filter
            if itm[0] != 'none':
                fxy_df = self.filter_data(gxy_df, itm[0], self.config)
            else:
                fxy_df = gxy_df

            #too few samples, trial should be passed
            if fxy_df is None:
                print(self.console['orange'] + 'too few samples' + self.console['ENDC'])
                fxy_df = gxy_df
                self.config['too_few_samples'] = True

            fxy_df = fxy_df.reset_index(drop=True)

            #------------------------------------------------------------fixation classification techniques
            #store fixations from eyelink
            if (self.config['source'] == 'eyelink'):
                eyelink_c_xy = self.cgxy

            #use eyelink calculated fixations if config.classify_eyelink = False
            if ((self.config['source'] == 'eyelink') and (self.config['classify_eyelink_data'] == False)):
                c_xy = self.cgxy
                ctype = 'eyelink'

            if ctype == 'idt':
                # if eyelink and using original fixations
                if ((self.config['source'] == 'eyelink') and (self.config['classify_eyelink_data'] == False)):
                  pass
                else:
                    c_xy = self.classify(self, df=fxy_df, dr_th=dr_th, di_th=di_th, ctype=ctype)

            if ctype == 'ivt':
                # if eyelink and using original fixations
                if ((self.config['source'] == 'eyelink') and (self.config['classify_eyelink_data'] == False)):
                  pass
                else:
                    c_xy = self.classify(self, self.config, fxy_df, v_th=v_th, ctype=ctype)

            if ctype == 'hmm':
                # if eyelink and using original fixations
                if ((self.config['source'] == 'eyelink') and (self.config['classify_eyelink_data'] == False)):
                  pass
                # if if samples are too few
                elif (self.config['too_few_samples']):
                  pass
                else:
                    dfc_xy = self.classify(self.config, df=fxy_df, ctype=ctype, filter_type=itm[1])
                    fxy_df = dfc_xy[0]
                    c_xy = dfc_xy[1]
                    del dfc_xy

            elif ctype == 'simple':
                # if eyelink and using original fixations
                if ((self.config['source'] == 'eyelink') and (self.config['classify_eyelink_data'] == False)):
                  pass
                else:
                    c_xy = self.classify(self.config, fxy_df, ctype=ctype,missing=_missing, maxdist=_maxdist, mindur=_mindur)
                    c_xy = pd.DataFrame(c_xy[1], columns=['start', 'end', 'duration', 'cx', 'cy'])

            #print eyelink orignal fixations
            if (self.config['source'] == 'eyelink'):
                print(self.console['green'] + 'eyelink fixations' + self.console['ENDC'])
                print(eyelink_c_xy)

            #print calculated fixations if from webgazer or calculating eyelink data
            if ((self.config['source'] == 'webgazer') or
                    (self.config['source'] == 'eyelink') and (self.config['classify_eyelink_data'] == True)):
                print(self.console['orange'] + 'calculated fixations' + self.console['ENDC'])
                #print(c_xy)

            #append filted data
            #_fgxy_df = append_data(gxy_df, fxy_df, itm[1])
            l_cxy.append(c_xy)
            if itm[0] != 'none':
                f_x = "%s_x" % (itm[1])
                f_y = "%s_y" % (itm[1])
            else:
                f_x = 'x'
                f_y = 'y'

            #class variable is created from hmm classification
            if (ctype == 'hmm') and (self.config['too_few_samples'] == False)\
                    and (set(["%s_class" % (itm[1])]).issubset(fxy_df.columns)):
                fxy_df = fxy_df[['x', 'y', "%s_class" % (itm[1])]]
            else:
                fxy_df["%s_class" % (itm[1])] = None
                fxy_df = fxy_df[['x', 'y', "%s_class" % (itm[1])]]

            #rename and append filtered x and y-axis
            fxy_df = fxy_df.rename(index=str, columns={"x": f_x, "y": f_y})
            l_gf.append(fxy_df)

        #combine list of all filtered and raw data
        for indx, item in enumerate(l_gf):
            l_gf[indx].index = range(len(l_gf[indx].index))
        fgxy_df = pd.concat(l_gf, axis=1, join_axes=[l_gf[0].index])

        return l_cxy, fgxy_df

    def append_classify(self, df, cg_df):
        """Appending classification to Dataframe.

        Parameters
        ----------
        df : :obj:`list`
            Pandas dataframe of raw data.
        gxy_df : :class:`pandas.DataFrame`
            Pandas dataframe of raw data of classification events.
        """
        for index, item in enumerate(cg_df):
            count = self.filters[index][1] + '_FID'
            for idx, rw in item.iterrows():
                start = rw['start']
                end = rw['end']

                #interval
                df.loc[df['timestamp'].between(start, end, inclusive=True), count] = idx

        return df

    def run(self, path, task_type="eyetracking", single_subject=False, single_trial=False, subject=0, trial=0, isMultiprocessing=True, cores=1):
        """Processing of data. Steps here include: cleaning data, fixation identification, and exporting data.

        Parameters
        ----------
        path : :obj:`string`
            Path of raw data.
        task_type : :obj:`string`
            Running analysis on `eyetracking` or `behavioral` data.
        single_subject : :obj:`bool`
            Whether to run function with all or single subject.
        single_trial : :obj:`bool`
            Whether to run function with all or single trial.
        subject : :obj:`int`
            Subject number. Only if single_subject = True.
        trial : :obj:`int`
            Trial number. Only if single_trial = True.
        isMultiprocessing : :obj:`bool`
            Whether multiprocessing of data will be used. Only if single_subject = False.
        cores : :obj:`int`
            Number of cores to use for multiprocessing. Only if single_subject = False & isMultiprocessing=True.
        
        Attributes
        ----------
        process : :obj:`bool`
            Process all data for export.
        """

        #----set config
        #set path
        self.config['path'] = path + '/' + self.config['task']
        # single_subject
        self.config['single_subject'] = single_subject
        # is single_trial
        self.config['single_trial'] = single_trial
        # subject and trial number
        self.config['subject'] = subject
        self.config['trial'] = trial
        # classification type
        ctype = self.config['ctype']
        # classification parameters
        missing=self.config['missing']
        maxdist=self.config['maxdist']
        mindur=self.config['mindur']
        v_th=self.config['v_th']
        dr_th=self.config['dr_th']
        di_th=self.config['di_th']

        #single subject
        if (self.config['single_subject']):
            print(self.console['orange'] + 'start-----------------' + self.console['ENDC'])
            print(self.console['orange'] + 'single_subject = %s, single_trial = %s'%(single_subject, single_trial) + self.console['ENDC'])
            print(self.console['orange'] + 'importing raw data' + self.console['ENDC'])
            #get core number and set as global variable
            self.thisCore = 0
            #get logger and set as global variable
            self.log=logging.getLogger(__name__)
            
            #get data
            gxy_df, df_path = self.getData()

            #check if eyetracking, else go to next subject
            isEyetracking = gxy_df['type'][0]
            if isEyetracking != 'eyetracking':
                print(self.console['orange'] + 'finished-----------------' + self.console['ENDC'])
                return

            #drop practice and sort
            gxy_df = gxy_df.drop(gxy_df[(gxy_df['event']=='Prac')].index)
            gxy_df = gxy_df.sort_values(['TrialNum','timestamp','sampleNum'], ascending=[True, True, True]).reset_index(drop=True)

            #if single trial
            if self.config['single_trial']:
                l_trials = [trial]
            else:
                #get list of all possible trials to pull data from
                l_trials = gxy_df['TrialNum'].unique()
        
                #remove nan from behavioral list (nan occurs when row trial_type == instructions)
                if (task_type == "behavioral"):
                    l_trials = [x for x in l_trials if not np.isnan(x)]

            #------------------------------------------------for each trial
            l_fgxy = []
            #start
            for idx in l_trials:
                #reset blocker for too few trials in filtering
                self.config['too_few_samples'] = False

                #set config trial number
                self.config['trial'] = idx
                print(self.console['orange'] + 'subject: %s'%(self.config['subject']) + self.console['ENDC'])
                print(self.console['orange'] + 'session: %s'%(self.config['session']) + self.console['ENDC'])
                print(self.console['orange'] + 'trial: %s'%(self.config['trial']) + self.console['ENDC'])

                #filter data to single trial
                df = gxy_df[gxy_df['TrialNum'] == idx].reset_index(drop=True)

                #get isCongruent
                if df['isCongruent'][0]:
                    self.config['isCongruent'] = 'Congruent'
                else:
                    self.config['isCongruent'] = 'Incongruent'

                #get emotion of left and right stimuli
                self.config['emotion'] = [df['left_mood'][0], df['right_mood'][0]]

                #1. preprocess data
                #get monitorSize
                monitorSize = (df['monitorSize.px'][0]).split('x')
                ##remove devicePixelRatio from monitor size
                monitorSize[0] = float(monitorSize[0]) / df['devicePixelRatio'][0]           
                monitorSize[1] = float(monitorSize[1]) / df['devicePixelRatio'][0]
                #store
                monitorSize = list(map(int, monitorSize))
                self.config['resolution.px'] = monitorSize
                #reset and store previous version as 'monitorSize_old'
                df['monitorSize_old'] = df['monitorSize.px']
                df['monitorSize.px'] = '%sx%s'%(monitorSize[0],monitorSize[1])
                
                #get scale
                ##scale container and stimulus if x-resolution is smaller than 1400
                if monitorSize[0] < 1400:
                    scaleImage = monitorSize[0]/1400
                ##dont scale
                else:
                    scaleImage = 1
                self.config['scaleImage'] = scaleImage
                ##preprocess
                df = self.preprocess(df, monitorSize)

                #get eyelink fixations
                print(self.console['blue'] + 'importing eyelink fixations' + self.console['ENDC'])
                if (self.config['source'] == 'eyelink'):
                    self.cgxy = self.eyelink_classify()

                #process data
                print(self.console['blue'] + 'processing data' + self.console['ENDC'])
                cgxy_df, fgxy_df = self.process(monitorSize, self.filters, df, self.config['trial'],
                                                _classify=self.config['classify'], ctype=ctype,
                                               _missing=missing, _maxdist=maxdist, _mindur=mindur,
                                               v_th=v_th, dr_th=dr_th, di_th=di_th)       

                #append classify to dataframe
                if self.config['classify']:
                    if self.config['ctype'] != 'hmm':
                        fgxy_df = self.append_classify(fgxy_df, cgxy_df)
                    #apply bounds on fixations
                    else:
                        fgxy_df=self.roi(filters=self.filters, flt=self.filters[0][1], df=fgxy_df)

                #check if all samples are flagged as fixations
                flt = self.filters[0][1]
                samples_fix_err = fgxy_df[fgxy_df["%s_fix_all"%(flt)]==True].count()["%s_fix_all"%(flt)]
                samples = len(fgxy_df)
                if samples_fix_err == samples:
                    fgxy_df['samples_fix_err'] = True
                else:
                    fgxy_df['samples_fix_err'] = False

                #sort
                fgxy_df = fgxy_df.sort_values(['TrialNum','sampleNum'],ascending=[True, True]).reset_index(drop=True)

                #append to ltrials, if there are at least 20 samples
                #this will be used to rebuild new dataframe for subject
                if (fgxy_df.shape[0] >= 1):
                    print(self.console['green'] + 'Preprocessing: trials with at least 20 samples' + self.console['ENDC'])
                    l_fgxy.append(fgxy_df)

            #combine list of dataframes into new dataframe
            df = pd.concat(l_fgxy, sort=False, ignore_index=True)
            df = df.sort_values(['TrialNum','sampleNum'],ascending=[True, True]).reset_index(drop=True)

            #save data
            print(self.console['blue'] + 'saving data' + self.console['ENDC'])
            subject = int(self.config['subject'])
            session = self.config['session']
            if self.config['save_data']:
                if self.config['single_trial']:
                    f_path = self.config['path']['processed'] + '/data/' + self.config['type'] + '/%s_%s_%s.csv'%(subject,session,trial)
                else:
                    f_path = self.config['path']['processed'] + '/data/' + self.config['type'] + '/%s_%s.csv'%(subject,session)
                #save
                df.to_csv(f_path, index=False)

            #finish
            return l_fgxy, cgxy_df, fgxy_df
        
        #all subjects
        elif (not self.config['single_subject']):
            print(self.console['orange'] + 'start-----------------' + self.console['ENDC'])
            print(self.console['orange'] + 'single_subject = %s, single_trial = %s'%(single_subject, single_trial) + self.console['ENDC'])
            print(self.console['blue'] + 'importing raw data' + self.console['ENDC'])
            #--------------------------------------------------for each subject
            def all_subjects(fdir, core):
                try:        
                    #for each file
                    for sbj in fdir:
                        print(self.console['blue'] + 'subject: %s'%(sbj) + self.console['ENDC'])
                        #set subject name, session
                        p = Path(sbj)
                        _subject, _session = (p.name.replace('abc', '').replace('.csv', '')).split("_", 1)
                        self.config['subject'] = _subject
                        self.config['session'] = _session
    
                        #get data
                        gxy_df, df_path = self.getData(path=sbj)
    
                        #if looking at eyetracking data
                        if (task_type == "eyetracking"):
                            #check if behavioral, else skip and go to next subject
                            isEyetracking = gxy_df['type'][0]
                            if isEyetracking != 'eyetracking':
                                print(self.console['orange'] + 'finished subject' + self.console['ENDC'])
                                continue
                            else:
                                pass
                        #else looking at behavioral data
                        elif (task_type == "behavioral"):
                            #check if behavioral, else skip and go to next subject
                            isBehavioral= gxy_df['type'][0]
                            if isBehavioral != 'behavioral':
                                print(self.console['orange'] + 'finished subject' + self.console['ENDC'])
                                continue
                            else:
                                pass
    
                        #drop practice and sort
                        gxy_df = gxy_df.drop(gxy_df[(gxy_df['event']=='Prac')].index)
                        gxy_df = gxy_df.sort_values(['TrialNum','timestamp','sampleNum'],ascending=[True, True, True]).reset_index(drop=True)

                        #if single trial
                        if self.config['single_trial']:
                            l_trials = [trial]
                        #else all trials
                        else:
                            #get list of all possible trials to pull data from
                            l_trials = gxy_df['TrialNum'].unique()
        
                            #remove nan from behavioral list (nan occurs when row trial_type == instructions)
                            if (task_type == "behavioral"):
                                l_trials = [x for x in l_trials if not np.isnan(x)]
                        
                        #------------------------------------------------for each trial
                        l_fgxy = []
                        #start
                        for idx in l_trials:
                            #reset blocker for too few trials in filtering
                            self.config['too_few_samples'] = False
    
                            #set config trial number
                            self.config['trial'] = idx
                            print(self.console['orange'] + 'subject: %s'%(_subject) + self.console['ENDC'])
                            print(self.console['orange'] + 'session: %s'%(_session) + self.console['ENDC'])
                            print(self.console['orange'] + 'trial: %s'%(self.config['trial']) + self.console['ENDC'])
    
                            #filter data to single trial
                            df = gxy_df[gxy_df['TrialNum'] == idx].reset_index(drop=True)                   
    
                            #get isCongruent
                            if df['isCongruent'][0]:
                                self.config['isCongruent'] = 'Congruent'
                            else:
                                self.config['isCongruent'] = 'Incongruent'
    
                            #get emotion of left and right stimuli
                            self.config['emotion'] = [df['left_mood'][0], df['right_mood'][0]]
                           
                            #get monitorSize
                            monitorSize = (df['monitorSize.px'][0]).split('x')
                            ##remove devicePixelRatio from monitor size
                            monitorSize[0] = float(monitorSize[0]) / df['devicePixelRatio'][0]           
                            monitorSize[1] = float(monitorSize[1]) / df['devicePixelRatio'][0]
                            ##store
                            monitorSize = list(map(int, monitorSize))
                            self.config['resolution.px'] = monitorSize
                            #reset and store previous version as 'monitorSize_old'
                            df['monitorSize_old'] = df['monitorSize.px']
                            df['monitorSize.px'] = '%sx%s'%(monitorSize[0],monitorSize[1])
                            
                            #get scale
                            ##scale container and stimulus if x-resolution is smaller than 1400
                            if monitorSize[0] < 1400:
                                scaleImage = monitorSize[0]/1400
                            ##dont scale
                            else:
                                scaleImage = 1
                            self.config['scaleImage'] = scaleImage
                            
                            #1. preprocess data
                            ## only if eyetracking data
                            if (task_type == "eyetracking"):
                                ##preprocess
                                df = self.preprocess(df, monitorSize)
        
                                #get eyelink fixations
                                print(self.console['blue'] + 'importing eyelink fixations' + self.console['ENDC'])
                                if (self.config['source'] == 'eyelink'):
                                    self.cgxy = self.eyelink_classify()
    
                                #process data, if eyetracking
                                print(self.console['blue'] + 'processing data' + self.console['ENDC'])
                                cgxy_df, fgxy_df = self.process(monitorSize, self.filters, df, self.config['trial'],
                                                                _classify=self.config['classify'],
                                                               ctype=ctype,
                                                               draw_plot=self.config['draw_plot'],
                                                               draw_heatmap=self.config['draw_heatmap'],
                                                               draw_gazeplot=self.config['draw_gazeplot'],
                                                               draw_fixplot=self.config['draw_fixplot'],
                                                               _missing=missing, _maxdist=maxdist, _mindur=mindur,
                                                               v_th=v_th, dr_th=dr_th, di_th=di_th)
    
                                #append classify to dataframe
                                if self.config['classify']:
                                    if self.config['ctype'] != 'hmm':
                                        fgxy_df = self.append_classify(fgxy_df, cgxy_df)
                                    #apply bounds on fixations
                                    else:
                                        fgxy_df=self.roi(filters=self.filters, flt=self.filters[0][1], df=fgxy_df)
                                
                                #check if all samples are flagged as fixations
                                flt = self.filters[0][1]
                                samples_fix_err = fgxy_df[fgxy_df["%s_fix_all"%(flt)]==True].count()["%s_fix_all"%(flt)]
                                samples = len(fgxy_df)
                                if samples_fix_err == samples:
                                    fgxy_df['samples_fix_err'] = True
                                else:
                                    fgxy_df['samples_fix_err'] = False
    
                                #sort
                                fgxy_df = fgxy_df.sort_values(['TrialNum','sampleNum'],
                                                              ascending=[True, True]).reset_index(drop=True)
                            # else if behavioral data continue
                            elif (task_type == "behavioral"):
                                fgxy_df = df
    
                            #append to ltrials
                            #this will be used to rebuild new dataframe for subject
                            if (fgxy_df.shape[0] >= 1):
                                l_fgxy.append(fgxy_df)
    
                        #combine list of dataframes into new dataframe
                        df = pd.concat(l_fgxy, sort=False, ignore_index=True)
                        df = df.sort_values(['TrialNum','sampleNum'],ascending=[True, True])\
                                .reset_index(drop=True)
    
                        #save data
                        print(self.console['blue'] + 'saving data' + self.console['ENDC'])
                        subject = int(self.config['subject'])
                        session = self.config['session']
                        if self.config['save_data']:
                            if self.config['single_trial']:
                                f_path = self.config['path']['processed'] + '/data/' + self.config['type'] + '/%s_%s_%s.csv'%(subject,session,trial)
                            else:
                                f_path = self.config['path']['processed'] + '/data/' + self.config['type'] + '/%s_%s.csv'%(subject,session)
                            #save
                            df.to_csv(f_path, index=False)
                            
                #--------------------------------------------------------------------------------------------------end
                #if all_subjects fails, save to log            
                except Exception as e:
                    self.log.error(e, exc_info=True)
                    
            #--------------------------------------------------prepare
            #multithreading
            #list of behavioral and eyetracking

            import multiprocessing
            
            #prepare collecting arguements
            arg = []

            #get directory
            fdir = glob.glob(os.path.join(self.config['path'] + "/*.csv"))
            
            #get cpu cores to be used
            cores = self.config['cores']

            #if requested cores is 1, run without multiprocessing
            if (cores == 1):
                isMultiprocessing = False
            ##if requested cores are less than/equal 7, and less than available cores plus 1
            elif (cores <= 7) and (multiprocessing.cpu_count() >= cores + 1):
                isMultiprocessing = True
                fdir_chunk = np.array_split(fdir, cores)
                for index in range(0, cores):
                    arg.append((fdir_chunk[index],index))
            ##else use less than half of total available cores
            else:
                isMultiprocessing = True
                cores = int(self.config['cores']/2)
                fdir_chunk = np.array_split(fdir, cores)
            #breakpoint()

            #------------------------------------------run multiprocessing
            #if not multiprocessing
            if not isMultiprocessing:
                all_subjects(fdir, cores)
            #else multiprocessing
            else:
                proc = [multiprocessing.Process(target=all_subjects,\
                        args=(fdir_chunk[x].tolist(), x)) for x in range(cores)]
                for p in proc:
                    p.daemon = True
                    p.start()

            #-----------------------------------------finished
            return cores, arg, proc

    def subject_metadata(self, fpath, spath):
        """
        Collect all subjects metadata.
        
        Parameters
        ----------
        fpath : :obj:`str`
            The directory path of all participant data.
        spath : :obj:`str`
            The directory path of all participant data.
            
        Returns
        -------
        df : :class:`ndarray`
            Pandas dataframe of subject metadata.
        """
        
        #----for timestamp
        _t0 = datetime.datetime.now()
        _f = self.debug(message='t', source="timestamp")
        print(self.console['blue'] + 'running metadata.summary()' + self.console['ENDC'])
        
        #----get directory
        fdir = glob.glob(fpath + "/*.csv")
        
        #----store metadata
        l_sub = []
        
        #----for each subject
        for index, sbj in enumerate(fdir):
            #read csv
            df = pd.read_csv(sbj)
            
            #add filename to df
            df['file'] = sbj
            
            #collect metadata for output by getting first row
            l_sub.append(df[df['trial_type'] == 'dotprobe-task'].iloc[0])

        #----convert list to dataframe
        df = pd.concat(l_sub, axis=1, keys=[s.name for s in l_sub], sort=False).T.reset_index(drop=True)
        
        #----format metadata and save
        df = Metadata.summary(df=df, path=spath)
                
        #----end
        print(self.console['blue']+'%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000))+self.console['ENDC'])
        
        return df
           
    def variables(self,df):
        """Output list of variables for easy html viewing.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Pandas dataframe of raw data. This is used as a filter to prevent unused participants from being included in the data.
        path : :obj:`str`
            The directory path save and read the hdf5 dataframe.

        Returns
        -------
        df_definitions : :class:`pandas.DataFrame`
        """
        #blank df for appending
        df_variable = pd.DataFrame()
        
        source = {
            'bias': ['init_gaze_bias','final_gaze_bias','gaze_bias','n_gaze_valid','dp_bias','var_dp_bias','n_dp_valid'],
            'demographic':['race','is_normalvision','is_student'],
            'behavioral': ['m_rt','accuracy','m_diff_stim','m_diff_dotloc'],
            'clinical': ['cesd_score','cesd_group','rrs_brooding'],
            'device': ['os','os_version','gpu','gpu_type','browser','browser_version', 'devicePixelRatio','monitorSize',
                       'windowSize','heap.used','heap.limit','WebcamMessage','webcamSize','webcam_brand','luminance',
                       'isPageVisible','is_calibrated','is_eyetracking','isFullscreen'],
            'other': ['nested']
        }
        
        #for each source
        for key, row in source.items():
            #convert to correct formats
            df_ = df[row].iloc[:2].loc[0,:].reset_index().rename(columns={'index':'variable', 0:'example'})
            
            #add column for definitions, type; reorganize
            df_['type'] = df[row].dtypes.to_frame().reset_index().rename(columns={0:'type'})['type']
            df_['group'] = key
            df_ = df_.loc[:,['variable','type','example','group']]
            
            #if key == behavioral, add row for trialnum, and trialnum_
            if key=='behavioral':
                df_.loc[-1] = ['TrialNum','int64',1,'behavioral']
                df_.loc[-1] = ['TrialNum_','int64',1,'behavioral']
                df_.loc[-1] = ['trialType','int64',1,'behavioral']
                df_.loc[-1] = ['trialType_','object','iaps','behavioral']
            
            #append
            df_variable = df_variable.append(df_)
            
        #reset index
        df_variable = df_variable.reset_index(level=0, drop=True)
             
        #import list of definitions and add to dataframe
        #to initially get list of variable to fill in definitions #df_variable.to_csv(definitions_path, index=None)
        
        #import definitions and merge to variables list
        definitions_path = self.config['path']['output'] + "/analysis/definitions.csv"
        df_definitions = pd.read_csv(definitions_path, float_precision='high')
        df_variable = pd.merge(df_variable, df_definitions, on='variable')
        
        #change order
        df_variable = df_variable[['variable','group','type','example','definition']]
        
        return df_variable
    
    def dwell(self, df, cores=1):
        """
        Calculate dwell time for sad and neutral images.

        Parameters
        ----------
        df : :class:`pandas.DataFrame`
            Pandas dataframe of raw data. This is used as a filter to prevent unused participants from being included in the data.
        cores : :class:`int`    
            Number of cores to use for multiprocessing.

        Returns
        -------
        df : :class:`pandas.DataFrame`
            Pandas dataframe with dwell time.
        error : :class:`list`
            List of participants that were not included in dataframe.
        """
        import multiprocessing, itertools
        
        #----for timestamp
        _t0 = datetime.datetime.now()
        _f = self.debug(message='t', source="timestamp")
        print(self.console['blue'] + 'running dwell()' + self.console['ENDC'])
        
        #----run
        def run(dir_=None, core=None, queue=None):
            print(self.console['green'] + 'processing.dwell.run(%s)'%(core) + self.console['ENDC'])
            dwell = []
            #----for each subject
            for index, row in dir_.iterrows():
                _path_ = row['path']
                _subject = row['participant']
                
                #----read csv as dataframe
                df_ = pd.read_csv(_path_, float_precision='high', low_memory=False)
                print('subject: %s; core: %s'%(_subject, core))
                
                #----if 198 trials continue else skip
                if (df_.drop_duplicates(subset="TrialNum", keep="first").shape[0] == 198):
                    #----for each trial
                    for _trial in range(0,198):
                        try:                            
                            #----keep relevant trial
                            df_0 = df_.loc[df_['TrialNum'] == _trial].reset_index(drop=True)
                            
                            #----get trialType 
                            _trialType = df_0['trialType'][0]
                        
                            #----drop columns
                            df_0 = df_0[['participant','TrialNum','timestamp','marker','dwell','trialType',
                                         'left_mood','right_mood','left_bound','right_bound']]
                            
                            #----get range between "stimulus onset" and "dotloc onset"
                            #get start and end markers
                            start_m = df_0[df_0['marker'] == "Stimulus Onset"].index.item()
                            end_m = df_0[df_0['marker'] == "Dotloc Onset"].index.item()
                            
                            #get dataframe
                            df_1 = df_0.iloc[start_m:end_m+1,:].reset_index(drop=True)
                            
                            #get difference between timestamp values
                            df_1['difference'] = df_1['timestamp'].shift(-1) - df_1['timestamp']              
                            
                            #----get location of sad/neutral images
                            if ((df_1['left_mood'][0]=='Sad') and (df_1['right_mood'][0]=='Neutral')):
                                sad = 'left_bound'
                                neutral = 'right_bound'
                            else:
                                neutral = 'left_bound'
                                sad = 'right_bound'
                            
                            #----get emotional dwell number of samples, if gaze is within dwell location                     
                            # nested by aoi (for within-group analysis)
                            ##neutral
                            df_neutral = df_1[(df_1[neutral]==True)]
                            ## subject, trial, aoi, dwell_num, dwell_time
                            dwell.append([_subject, _trial, _trialType, 'neutral', df_neutral.shape[0], df_neutral['difference'].sum()])
                            ##sad
                            df_sad = df_1[(df_1[sad]==True)]
                            ## subject, trial, aoi, dwell_num, dwell_time
                            dwell.append([_subject, _trial, _trialType, 'sad', df_sad.shape[0], df_sad['difference'].sum()])
                            
                        #if exception
                        except Exception as e:
                            line = sys.exc_info()[-1].tb_lineno
                            print('subject: %s; trial: %s; error: %s; line: %s'%(_subject, _trial, e, line))
                        
                else:
                    print('subject:%s; error:too few trials'%(_subject))
            
            #----add to multithreading queue
            if isMultiprocessing:
                #queue
                queue.put(dwell)
                
            if not isMultiprocessing:
                return dwell
        
        #----finished
        def finished(output, error=None):
            print(self.console['green'] + 'processing.dwell.finished()' + self.console['ENDC'])
            if isMultiprocessing:
                #----create output df
                df = pd.DataFrame(list(itertools.chain(*output)),
                                  columns=['participant','trial','trialType','aoi','dwell_num','dwell_time'])
                
                #----create error df
                #check if data came out correctly by looking at ampunt of trials and participants outputted
                error = df.groupby(['participant','aoi']).agg(['count'])
            else:
                #----create output df
                df = pd.DataFrame(output, columns=['participant','trial','trialType','aoi','dwell_num','dwell_time'])
                #check if data came out correctly by looking at ampunt of trials and participants outputted
                error = df.groupby(['participant','aoi']).agg(['count'])
            
            #----end
            #timestamp
            print(self.console['blue']+'%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000))+self.console['ENDC'])
            return df, error
        
        #--------------------get list of all participants
        ##get directory
        _dir = glob.glob(os.path.join(self.config['path']['processed'] + '/data/eyetracking' + "/*.csv"))
        _sbj_session = [(Path(x).name.replace('abc', '').replace('.csv', '')).split("_", 1) for x in _dir]
        dir_ = [[z,*x] for z,x in zip(_dir,_sbj_session)]
        dir_ = pd.DataFrame(dir_, columns=['path','participant','session'])

        #------------------------------------------check if running using multiprocessing
        #----get cores
        ##if requested cores is 1, run without multiprocessing
        if ((cores == 0) or (cores == 1)):
            isMultiprocessing = False
            print(self.console['green'] + 'processing.dwell() not multiprocessing' + self.console['ENDC'])
        ##else if requested cores are less than/equal 7, and less than available cores plus 1
        elif ((cores <= 7) and (multiprocessing.cpu_count() >= cores + 1)):
            isMultiprocessing = True
            dir_p = np.array_split(dir_, cores)
            print(self.console['green'] + 'processing.dwell() multiprocessing with %s cores'%(cores) + self.console['ENDC'])
        ##else use less than half of total available cores
        else:
            isMultiprocessing = True
            cores = int(cores/2)
            dir_p = np.array_split(_dir, cores)
            print(self.console['green'] + 'processing.dwell() multiprocessing with %s cores'%(cores) + self.console['ENDC'])
            
        #------------------------------------------multiprocessing
        #if not multiprocessing
        if not isMultiprocessing:
            #----start
            output = run(dir_, cores)
            
            #----after finished 
            df_dwell, df_error = finished(output=output)
            
            #----merge
            df_dwell["participant"] = df_dwell[["participant"]].astype(np.int64)
            df = df_dwell.merge((df[['cesd_group','participant']].drop_duplicates(subset="participant", keep="first")), on='participant')
            
            #----finished
            return df, df_error
        #else multiprocessing
        else:
            #collect each pipe (this is used to build send and recieve portions of output)
            queue = multiprocessing.Queue()
            
            #----collect each thread
            process = [multiprocessing.Process(target=run,args=(dir_p[core], core, queue,)) for core in range(cores)]
            
            #start each thread
            for p in process:
                p.daemon = True
                p.start()
            
            #return queues
            #note: see https://stackoverflow.com/a/45829852
            returns = []
            for p in process:
                returns.append(queue.get())
                
            #wait for each process to finish
            for p in process:
                p.join()

            #---- finished multiprocessing
            df_dwell, df_error = finished(output=returns)
            
            #----merge
            df_dwell["participant"] = df_dwell[["participant"]].astype(np.int64)
            df = df_dwell.merge((df[['cesd_group','participant']].drop_duplicates(subset="participant",keep="first")), on='participant')
            return df, df_error
    
    def onset_diff(self, df0, merge=None, cores=1):
        """Calculate differences in onset presentation (stimulus, dotloc) using bokeh, seaborn, and pandas.

        Parameters
        ----------
        df0 : :class:`pandas.DataFrame`
            Pandas dataframe of raw data. This is used to merge variables that may be useful for analysis.
        merge : :class:`list` or `None`
            Variables to merge into returned df.
        cores : :class:`int`    
            Number of cores to use for multiprocessing.

        Returns
        -------
        df1 : :class:`pandas.DataFrame`
            Pandas dataframe.
        error : :class:`pandas.DataFrame`
            Dataframe of each participants and the amount trials included in their data.
        drop : :class:`list`
            List of participants that are 3 SD from median.
        """
        import multiprocessing
        #timestamp
        _t0 = datetime.datetime.now()
        _f = self.debug(message='t', source="timestamp")
        
        #----run
        def run(dir_=None, core=None, queue=None):
            print(self.console['green'] + 'processing.onset_diff.run(%s)'%(core) + self.console['ENDC'])
            onset = []
            #----for each subject
            for index, row in dir_.iterrows():
                _path_ = row['path']
                _subject = row['participant']
                
                #----read csv as dataframe
                df = pd.read_csv(_path_, float_precision='high', low_memory=False)
                print('subject: %s; core: %s'%(_subject, core))
                
                #----if 198 trials continue else skip
                if (df.drop_duplicates(subset="TrialNum", keep="first").shape[0] == 198):  
                    #----format column
                    df['RT'] = df['RT'].astype('float64')
                    
                    #----calculate number of samples in trial
                    df['trial_samples'] = df.groupby('TrialNum')['type'].transform(len)
                    
                    #condense to subject:trial-level
                    df = df.loc[df.drop_duplicates(subset="TrialNum", keep="first").index].reset_index(level=0, drop=True)
    
                    #----calculate scores
                    # stimulus onset
                    df['diff_stim'] = df.apply(lambda x: abs(x['Stim_onset.t'] - 1500), axis=1)
                    # dotloc onset
                    df['diff_dotloc'] = df.apply(lambda x: abs(x['DotLoc_onset.t'] - (1500 + 4500))
                    if (x['trialType'] == 'iaps') else abs(x['DotLoc_onset.t'] - abs(1500 + 3000)), axis=1)
                    
                    #----relevant data only                                    
                    #keep important columns
                    df = df[['participant', 'TrialNum', 'diff_dotloc', 'diff_stim', 'Key_Resp.rt','Key_Resp.acc','trialType']]
                    #group by trial
                    
                    #store to list
                    onset.append(df)
                    
                else:
                    print('%s; error: too few trials'%(_subject))
            
            #----add to multithreading queue
            if isMultiprocessing:
                #queue
                queue.put(onset)
                
            if not isMultiprocessing:
                return onset
        
        #--------finished
        def finished(output, isMultiprocessing, error=None):
            #----concat list of dataframes
            if isMultiprocessing:
                #flatten
                output_ =  [itm for l in output for itm in l]
                df = pd.concat(output_)
            else:
                df = pd.concat(output)
            #rename
            df = df.rename(columns={'Key_Resp.rt':'Key_Resp_rt','Key_Resp.acc':'Key_Resp_acc'})
            
            #--get median onset error
            df['m_diff_dotloc'] = df.groupby(["participant"])['diff_dotloc'].transform('median')
            df['m_diff_stim'] = df.groupby(["participant"])['diff_stim'].transform('median')
            df['m_rt'] = df.groupby(["participant"])['Key_Resp_rt'].transform('median')
            #--get accuracy
            df['accuracy'] = df.groupby(["participant"])['Key_Resp_acc'].transform('sum')    
            #--transform trial number to 0-1 for analysis
            df['TrialNum_'] = df.groupby(["participant"])['TrialNum'].transform(lambda x: x / 198)

            #----store participants with diff_dotloc or diff_stim > 500
            #drop dotloc diff time 3 SD > mean
            ##diff_dotloc
            error_std = df['diff_dotloc'].std() * 3
            drop_1 = list(df.loc[df['diff_dotloc'] > error_std].drop_duplicates(subset="participant", keep="first")['participant'])
            ##diff_stim
            error_std = df['diff_stim'].std() * 3
            drop_2 = list(df.loc[df['diff_stim'] > error_std].drop_duplicates(subset="participant", keep="first")['participant'])
            drop = list(np.unique(drop_1 + drop_2))
            
            #create new column where if onset diff greater than 500msec, then True, else False
            df['onset>500'] = np.where(df['participant'].isin(drop), True, False)
            
            #check if data came out correctly by looking at ampunt of trials and participants outputted
            error = df.groupby(['participant']).agg(['count'])
            
            #merge with df
            if merge is not None:
                df["participant"] = df[["participant"]].astype(np.int64)
                df1 = df.merge((df0[merge].drop_duplicates(subset="participant", keep="first")), on='participant')
                
            #check differences
            l_df1 = df1['participant'].unique().tolist()
            l_df = df['participant'].unique().tolist()
            missing = set(l_df1) ^ set(l_df)
            
            #----end
            print(self.console['blue']+'%s finished in %s msec'%(_f,((datetime.datetime.now()-_t0).total_seconds()*1000))+self.console['ENDC'])
            return df1, error, drop
        #--------------------get list of all participants
        ##get directory
        _dir = glob.glob(os.path.join(self.config['path']['processed'] + '/data/eyetracking' + "/*.csv"))
        _sbj_session = [(Path(x).name.replace('abc', '').replace('.csv', '')).split("_", 1) for x in _dir]
        dir_ = [[z,*x] for z,x in zip(_dir,_sbj_session)]
        dir_ = pd.DataFrame(dir_, columns=['path','participant','session'])
        
        #------------------------------------------check if running using multiprocessing
        #----get cores
        ##if requested cores is 1, run without multiprocessing
        if ((cores == 0) or (cores == 1)):
            isMultiprocessing = False
            print(self.console['green'] + 'processing.onset_diff() not multiprocessing' + self.console['ENDC'])
        ##else if requested cores are less than/equal 7, and less than available cores plus 1
        elif ((cores <= 7) and (multiprocessing.cpu_count() >= cores + 1)):
            isMultiprocessing = True
            dir_p = np.array_split(dir_, cores)
            print(self.console['green'] + 'processing.onset_diff() multiprocessing with %s cores'%(cores) + self.console['ENDC'])
        ##else use less than half of total available cores
        else:
            isMultiprocessing = True
            cores = int(cores/2)
            dir_p = np.array_split(_dir, cores)
            print(self.console['green'] + 'processing.onset_diff() multiprocessing with %s cores'%(cores) + self.console['ENDC'])
            
        #------------------------------------------multiprocessing
        #if not multiprocessing
        if not isMultiprocessing:
            #----start
            output = run(dir_, cores)
            
            #----after finished 
            df_onset, df_error, drop = finished(isMultiprocessing=isMultiprocessing, output=output)
            
            return df_onset, df_error, drop
        #else multiprocessing
        else:
            #collect each pipe (this is used to build send and recieve portions of output)
            #pipes = [multiprocessing.Pipe(False) for pipe in range(cores)]
            queue = multiprocessing.Queue()
            
            #----collect each thread
            process = [multiprocessing.Process(target=run,args=(dir_p[core], core, queue,)) for core in range(cores)]
            
            #start each thread
            for p in process:
                p.daemon = True
                p.start()
            
            #return queues
            #note: see https://stackoverflow.com/a/45829852
            returns = []
            for p in process:
                #queue.get() will block #https://stackoverflow.com/a/45829852
                returns.append(queue.get())
                
            #wait for each process to finish
            for p in process:
                p.join()

            #---- finished multiprocessing
            df_onset, df_error, drop = finished(isMultiprocessing=isMultiprocessing, output=returns)
            
            return df_onset, df_error, drop


#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
| `@purpose`: Module designed for working with eyetracking data.  
| `@date`: Created on Sat May 1 15:12:38 2019  
| `@author`: Semeon Risom  
| `@email`: semeon.risom@gmail.com  
| `@url`: https://semeon.io/d/mdl
"""
# allowed imports
__all__ = ['Calibration','Eyelink','ROI','pylink']

# core
from pdb import set_trace as breakpoint
import os
import sys

# relative paths
sys.path.append(os.path.dirname(os.path.realpath(__file__)))

# imports
import pylink
from ..settings import copyInherit
from ._roi import ROI as _ROI
from ._eyelink import Eyelink as _Eyelink
from ._calibration import Calibration as _Calibration

# set as namespace package
# see https://stackoverflow.com/questions/41621131/python-namespace-packages-in-python3
try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
    del pkg_resources
except ImportError:
    import pkgutil
    __path__ = pkgutil.extend_path(__path__, __name__)
    del pkgutil

# classes
@copyInherit(_Eyelink)
class Eyelink(_Eyelink):
	"""Interface for the SR Research Eyelink eyetracking system."""
	def __init__(self, window, timer, isPsychopy=True, subject=None, **kwargs):
		super().__init__(window, timer, isPsychopy=True, subject=None, **kwargs)

@copyInherit(_Calibration)
class Calibration(_Calibration):
	"""Allow mdl.eyetracking.Eyelink to initiate calibration/validation/drift correction."""
	def __init__(self, w, h, tracker, window):
		super().__init__(w, h, tracker, window)

@copyInherit(_ROI)
class ROI(_ROI):
	"""Generate regions of interest that can be used for data processing and analysis."""
	def __init__(self, image_path=None, output_path=None, metadata_path=None, shape='box', **kwargs):
		super().__init__(image_path=image_path, output_path=output_path, metadata_path=metadata_path, shape=shape, **kwargs)

# finished 
#del copyInherit, os, sys, _ROI, _Eyelink, _Calibration



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| `@purpose`: Module designed for working with the R33 study.  
| `@date`: Created on Sat May 1 15:12:38 2019  
| `@author`: Semeon Risom  
| `@email`: semeon.risom@gmail.com  
| `@url`: https://semeon.io/d/mdl
"""
# allowed imports
__all__ = ['Classify','Metadata','Model','nslr_hmm','Processing','Settings']

# core
from pdb import set_trace as breakpoint
import os
import sys

# relative paths
sys.path.append(os.path.dirname(os.path.realpath(__file__)))

# imports
from . import _nslr_hmm
from ..settings import copyInherit
from ._classify import Classify as _Classify
from ._metadata import Metadata as _Metadata
from ._model import Model as _Model
from ._processing import Processing as _Processing
from ._settings import Settings as _Settings

# set as namespace package
# see https://stackoverflow.com/questions/41621131/python-namespace-packages-in-python3
try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
    del pkg_resources
except ImportError:
    import pkgutil
    __path__ = pkgutil.extend_path(__path__, __name__)
    del pkgutil

# classes
@copyInherit(_Classify)
class Classify(_Classify):
	"""Classification of eyetracking data for mdl.r33.procesing."""
	def __init__(self, isLibrary=False):
		super().__init__(isLibrary=False)

@copyInherit(_Metadata)
class Metadata(_Metadata):
	"""Process participants metadata for analysis and export."""
	def __init__(self, isLibrary=False):
		super().__init__(isLibrary=False)

@copyInherit(_Model)
class Model(_Model):
	"""Run statistical models for analysis."""
	def __init__(self, isLibrary=False):
		super().__init__(isLibrary=False)

@copyInherit(_Processing)
class Processing(_Processing):
	"""Hub for running processing and analyzing raw data."""
	def __init__(self, config, isLibrary=False):
		super().__init__(config, isLibrary=False)

@copyInherit(_Settings)
class Settings(_Settings):
	"""Default settings for mdl.r33.Processing"""
	def __init__(self, isLibrary=False):
		super().__init__(isLibrary=False)

# finished
# del os, sys, _Classify, _Metadata, _Model, _Processing, _Settings



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
| @purpose: Download raw data from apache, Box, or REDCap servers.  
| @date: Created on Sat May 1 15:12:38 2019  
| @author: Semeon Risom  
| @email: semeon.risom@gmail.com  
| @url: https://semeon.io/d/mdl
"""

# available functions
__all__ = ['Download']

# core
from pdb import set_trace as breakpoint
import os
from datetime import datetime
from pathlib import Path

# local libraries
if __name__ == '__main__':
	from . import settings

# required external libraries
__required__ = ['paramiko','openpyxl','requests','ftplib','pandas']

class Download():
    """Download raw data from apache, Box, or REDCap servers."""
    def __init__(self, isLibrary=False):
        """Initiate the mdl.Download module.

        Parameters
        ----------
        isLibrary : :obj:`bool`
            Check if required libraries are available.
        """
        #check libraries
        if isLibrary:
            settings.library(__required__)

    @classmethod
    def REDCap(cls, path, token, url, content, payload=None, **kwargs):
        """Download data from an Research Electronic Data Capture (REDCap) server.

        Parameters
        ----------
        path : :obj:`str`
            Path to save data.
            For example::
                >>> path = '/Users/mdl-admin/Desktop/r33/redcap'
        token : :obj:`str`
            The API token specific to your REDCap project and username. This is usually found on the Applications > API page.
            For example::
                >>> token = 'D19859823032SFDMR24395298'
        url : :obj:`str`
            The URL of the REDCap project.
            For example::
                >>> url = 'https://redcap.prc.utexas.edu/redcap/api/'.
        content : :obj:`str` {report, file, raw, arm, instrument, returnFormat, metadata, project, surveyLink, user, participantList}
            Type of export. Examples include exporting a report (`report`), file (`file`), or project info (`project`).
        payload_ : :obj:`dict` or :obj:`None`, optional
            Manually submit parameters for exporting or importing data. Can be entered within the function for convenience.
        **kwargs : :obj:`str` or :obj:`None`, optional
            Additional properties, relevent for specific content types. Here's a list of available properties:

            .. list-table::
                :class: kwargs
                :widths: 25 50
                :header-rows: 1

                * - Property
                  - Description
                * - **report_id** : :obj:`str`
                  - (report, record) The report ID number provided next to the report name on the report list page.
                * - **cformat** : :obj:`str` {csv, json, xml, odm}
                  - Format to return data, either csv, json, xml, or odm. Default is json.
                * - **ctype** : :obj:`str`
                  - Shape of data. Default is flat.
                * - **rawOrLabel**: :obj:`str` {raw, label}
                  - (report, record) TExport the raw coded values or labels for the options of multiple choice fields.
                * - **rawOrLabelHeaders**: :obj:`str`
                  - (report, record) TExport the variable/field names (raw) or the field labels (label).
                * - **exportCheckboxLabel**: :obj:`str`
                  - Specifies the format of checkbox field values specifically when exporting the data as labels (i.e., when rawOrLabel=label).
                * - **returnFormat** : :obj:`str`
                  -  Format to return errors. Default is `json`.

        Returns
        -------
        log : :obj:`pandas.DataFrame` or :obj:`None`
            Pandas dataframe of each download request.
        content : :obj:`pandas.DataFrame` or :obj:`None`
            Pandas dataframe of all data downloaded.
        start, end : :obj:`str`
            Timestamp (ISO format) and name of most recent (`end`) and first (`start`) file created in folder.
        now : :obj:`str`
            Current timestamp in ISO format.

        """
        import openpyxl, requests
        import pandas as pd

        #----constants, lists to prepare
        ldate = [] #list of dates
        file_num = 0 #file counter
        # bool
        log = None #log of events
        start = None #most recent file
        end = None #most recent file

        #----path
        # log
        name = Path(path).name
        log_path = os.path.abspath(os.path.dirname(path + "/../"))
		
        # destination
        path = Path(path)

        #----kwargs
        # general
        cformat = kwargs['cformat'] if "cformat" in kwargs else 'json'
        ctype = kwargs['ctype'] if "ctype" in kwargs else 'flat'
        rawOrLabel = kwargs['rawOrLabel'] if "rawOrLabel" in kwargs else 'raw'
        rawOrLabelHeaders = kwargs['rawOrLabelHeaders'] if "rawOrLabelHeaders" in kwargs else 'raw'
        exportCheckboxLabel = kwargs['exportCheckboxLabel'] if "exportCheckboxLabel" in kwargs else 'false'
        returnFormat = kwargs['returnFormat'] if "returnFormat" in kwargs else 'json'
        # report
        report_id = kwargs['report_id'] if "report_id" in kwargs else 'None'
        # participantList
        instrument = kwargs['instrument'] if "instrument" in kwargs else 'cssrs'
        event = kwargs['event'] if "event" in kwargs else 'online_eligibility_arm_1'

        #----start
        settings.console('connecting to REDCap', 'blue')

        #----make sure local path exists
        settings.console('local folder: %s'%(Path(path)), 'blue')
        if not os.path.exists(Path(path)):
            settings.console('creating local folder: %s'%(Path(path)), 'blue')
            os.makedirs(Path(path))

        #----start download
        # prepare
        if payload == None:
            # report
            if content == 'report':
                payload = {
                    'token': token,
                    'content': content,
                    'format': cformat,
                    'type': ctype,
                    'returnFormat': returnFormat,
                    'report_id': report_id,
                    'rawOrLabel': rawOrLabel,
                    'rawOrLabelHeaders': rawOrLabelHeaders,
                    'exportCheckboxLabel': exportCheckboxLabel
                }
            elif content == 'participantList':
                payload = {
                    'token': token,
                    'content': content,
                    'format': cformat,
                    'returnFormat': returnFormat,
                    'instrument': instrument,
                    'event': event,
                }
            elif content == 'metadata':
                payload = {
                    'token': token,
                    'content': content,
                    'format': cformat,
                    'returnFormat': returnFormat,
                }
            elif content == 'project':
                payload = {
                    'token': token,
                    'content': content,
                    'format': cformat,
                    'returnFormat': returnFormat,
                }
            else:
                raise Exception('Not finished. Content `%s` unavailable.'%(content))

        # get data
        response = requests.post(url, data=payload)
        json = response.json()
        # if error
        if 'error' in json:
            raise Exception('Export failed. Reason: %s.'%(json['error']))

        # convert to dataframe
        df = pd.DataFrame(json)

        # save data
        path_= os.path.dirname(path) + "/REDCap/%s.xlsx"%(content)
        df.to_excel(path_, index=False)

        # if list isnt empty get start and end dates
        if ldate:
            ## start
            start = min(map(lambda x: [x[0],x[1]], ldate))
            settings.console('oldest file: %s'%(start[0]), 'blue')
            ## end
            end = max(map(lambda x: [x[0],x[1]], ldate))
            settings.console('newest file: %s'%(end[0]), 'blue')

        #----log
        # if new files
        now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        #!!! add values to go into dataframe----
        row = [df.shape[0],token,content,now]
        # log file path
        fpath = "%s/%s.xlsx"%(log_path, name)

        # if log exists, update
        if os.path.exists(fpath):
            settings.console("Log updated @: %s"%(fpath), 'blue')
            #load file
            wb = openpyxl.load_workbook(fpath)
            # Select First Worksheet
            ws = wb.worksheets[0]
            # add data
            ws.append(row)
            # update
            wb.save(fpath)
            # import log
            log = pd.ExcelFile(fpath)
        # else create log
        else:
            settings.console('Log created @: %s'%(fpath), 'blue')
            headers = ['subjects', 'token', 'content', 'date']
            # creating datafame and save as xlsx
            log = pd.DataFrame([row], columns=headers)
            log.to_excel(fpath, index=False)

        return log, start, end, now

    @classmethod
    def SFTP(cls, source, destination, hostname, username, password, **kwargs):
        """Connect to a remote server using a Secure File Transfer Protocol (SFTP).

        Parameters
        ----------
        source : :obj:`str`
            The directory path to retrieve paticipant data.
        destination : :obj:`str`
            The directory path to save paticipant data.
        hostname : :obj:`str`
            SFTP hostname.
        username : :obj:`str`
            SFTP username.
        password : :obj:`str`
            SFTP password.
        **kwargs : :obj:`str` or :obj:`None`, optional
            Additional properties, relevent for specific content types. Here's a list of available properties:

            .. list-table::
                :class: kwargs
                :widths: 25 50
                :header-rows: 1

                * - Property
                  - Description
                * - **filetype** : ::obj:`str` or :obj:`None`
                  - Filetype to download. Default is csv.

        Returns
        -------
        log : :obj:`pandas.DataFrame` or :obj:`None`
            Pandas dataframe of each download request.
        content : :obj:`pandas.DataFrame` or :obj:`None`
            Pandas dataframe of all files downloaded.
        start, end : :obj:`str`
            Timestamp (ISO format) and name of most recent (`end`) and first (`start`) file created in folder.
        now : :obj:`str`
            Current timestamp in ISO format.

        Examples
        --------
        >>> name='r33'; source='/home/utweb/utw1211/public_html/r33'; d='/Users/mdl/Desktop/r33/'; un='utw1211'; pwd='43#!9amZ?K$'
        >>> log, start, end, now = download.SFTP(source=s, destination=d, hostname=hostname, username=un, password=pwd)

        """
        import paramiko, openpyxl
        import pandas as pd

        #----kwargs
        lfiletype = kwargs["filetype"].replace('.','') if "filetype" in kwargs else ['csv']

        #----constants, lists to prepare
        ldate = [] #list of dates
        file_num = 0 #file counter
        # bool
        log = None #log of events
        start = None #most recent file
        end = None #most recent file
        # path

        ## log
        name = Path(destination).name
        log_path = os.path.abspath(os.path.dirname(destination + "/../"))
        ## destination
        destination = Path(destination)

        settings.console('connecting to sftp', 'blue')

        #----make sure local path exists
        settings.console('local folder: %s'%(Path(destination)), 'blue')
        if not os.path.exists(Path(destination)):
            settings.console('creating local folder: %s'%(Path(destination)), 'blue')
            os.makedirs(Path(destination))

        #----connect to SSH client
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(hostname=hostname, username=username, password=password)
        sftp = client.open_sftp()

        #----set path and get list of files
        sftp.chdir(source)
        remote_path = str(sftp.getcwd()) + '/'
        remote_directory = sftp.listdir()
        remote_attr = sftp.listdir_attr()

        #----check number of files in remote directory
        # list comprehension: <for each file in remote directory> <check if file extension matches> <and if file extension isn't empty string>
        remote_num = len([x for x in remote_directory if Path(x).suffix.replace('.','') in lfiletype and bool(Path(x).suffix.replace('.',''))])
        try:
            if remote_num >= 1:
                settings.console('Total files in %s: %s'%(remote_path, remote_num), 'blue')
            else:
                raise Exception("Error: No %s found in folder: '%s'"%(lfiletype, remote_path))
        except Exception as error:
            settings.console(str(error), 'red')
            raise

        #----export files
        settings.console("Starting download from %s"%(remote_path), 'blue')
        # for each file in remote directory
        for file in remote_attr:
            # file
            filename = file.filename
            filetype = Path(filename).suffix.replace('.','')

            # if file is correct filetype and not empty string
            if (filetype in lfiletype) and (bool(filetype)):
                # path
                remote_fpath = remote_path + filename
                local_fpath = '%s/%s'%(destination, filename)
                # date
                date = datetime.fromtimestamp(file.st_mtime).strftime('%Y-%m-%d %H:%M:%S')
                ldate.append([date, filename])

                # if file not in local drive, download
                if not os.path.isfile(local_fpath):
                    settings.console('filename: %s, type: %s, date: %s'%(filename, lfiletype, date), 'green')
                    # counter
                    file_num = file_num + 1
                    # download
                    sftp.get(remote_fpath, local_fpath)

        # if list isnt empty get start and end dates
        if ldate:
            ## start
            start = min(map(lambda x: [x[0],x[1]], ldate))
            settings.console('oldest file: %s'%(start[0]), 'blue')
            ## end
            end = max(map(lambda x: [x[0],x[1]], ldate))
            settings.console('newest file: %s'%(end[0]), 'blue')

        #closing sftp
        sftp.close()

        #----log
        # if new files
        now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        if file_num > 0:
            settings.console("Finished downloading files", 'blue')
            row = [now, file_num, remote_num, end[1]]
            fpath = "%s/%s.xlsx"%(log_path, name)
            # if log exists, update
            if os.path.exists(fpath):
                settings.console("Log updated @: %s"%(fpath), 'blue')
                #load file
                wb = openpyxl.load_workbook(fpath)
                # Select First Worksheet
                ws = wb.worksheets[0]
                # add data
                ws.append(row)
                # update
                wb.save(fpath)
                # import log
                log = pd.ExcelFile(fpath)
            # else create log
            else:
                settings.console('Log created @: %s'%(fpath), 'blue')
                headers = ['folder', 'downloaded', 'total', 'newest']
                # creating datafame and save as xlsx
                log = pd.DataFrame([row], columns=headers)
                log.to_excel(fpath, index=False)
        else:
            settings.console("No files to download", 'blue')

        return log, start, end, now


    @classmethod
    def box(cls, source, destination, server, username, password):
        """Connect to Box cloud cloud storage service, using File Transfer Protocol over SSL (FTPS).

        Parameters
        ----------
        source : :obj:`str`
            The remoate path on box to retrieve data.
        destination : :obj:`str`
            The local path to download data.
        server : :obj:`str`
            Name of box server.
        username : :obj:`str`
            Box account email address.
        password : :obj:`str`
            Box account password.

        Notes
        -----
        - The username and password must be seperate from your 'Secure Sign On' associated with your UT login.
        - box only allows connections via FTPS for downloading and uploading data.
        - For more information: https://community.box.com/t5/Upload-and-Download-Files-and/Using-Box-with-FTP-or-FTPS/ta-p/26050

        """
        import openpyxl
        import pandas as pd
        import ftplib

        #----constants, lists to prepare
        ldate = [] #list of dates
        file_num = 0 #file counter
        # bool
        log = None #log of events
        start = None #most recent file
        end = None #most recent file
        # path

        ## log
        name = Path(destination).name
        log_path = os.path.abspath(os.path.dirname(destination)) + "/%s.xlsx"%(name)
        ## destination
        destination = Path(destination)

        settings.console('connecting to sftp', 'blue')

        #----make sure local path exists
        settings.console('local folder: %s'%(Path(destination)), 'blue')
        if not os.path.exists(Path(destination)):
            settings.console('creating local folder: %s'%(Path(destination)), 'blue')
            os.makedirs(Path(destination))

        #----start ftplib
        ftps = ftplib.FTP_TLS()
        ftps.set_debuglevel(2)

        # connect to host and authorize
        ftps.connect(host=server, port=990)
        ftps.set_pasv(True)
        ftps.auth()
        ftps.prot_p()

        # login
        ftps.login(username, password)

        #check path
        path = ftps.pwd()

        #get files in directory
        files = ftps.dir()

        # file to send
        # file = open('kitten.jpg','rb')
        # send the file
        # session.storbinary('STOR kitten.jpg', file)

        # logout
        ftps.quit()

        log = ''
        start = ''
        end = ''
        now = ''

        return log, start, end, now



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
| @purpose: Default settings for processing.py
| @date: Created on Sat May 1 15:12:38 2019
| @author: Semeon Risom
| @email: semeon.risom@gmail.com
| @url: https://semeon.io/d/mdl  
"""

from pdb import set_trace as breakpoint
from datetime import datetime
import os
import re
import sys

# available functions
__all__ = ['console','link','popover','stn','library','path']

# path
path = {
	'home': os.path.dirname(os.path.abspath(__file__))
}

def copyInherit(fromfunc, sep="\n"):
	"""Copy the docstring of inherited functions and classes.

	Parameters
	----------
	fromfunc : [type]
		[description]
	sep : str, optional
		[description], by default "\n"

	Returns
	-------
	[type]
		[description]
	"""
	#breakpoint()
	def _decorator(func):
		sourcedoc = fromfunc.__doc__
		if func.__doc__ == None:
			func.__doc__ = sourcedoc
		return func

	return _decorator

def time():
	"""
	Get local time in ISO-8601 format.

	Returns
	-------
	iso : :class:`str`
		ISO-8601 datetime format, with timezone.

	Example
	-------
	>>> __time__()
	'2019-04-23 11:29:44-05:00'
	"""

	iso = datetime.now().astimezone().replace(microsecond=0).isoformat()

	return iso

def console(message, color='blue'):
	"""
	Allows user-friend console logging using ANSI escape codes.

	Parameters
	----------
	message : :class:`str`
		Message to send to console.
	color : :class:`str`
		Name of color or ANSI escape event to use.

	Returns
	-------
	result : :class:`str`
		ANSI escape code.

	Examples
	--------
	>>> console(self, message, color='blue)

	Notes
	-----
	Colors are produced using ASCII Oct format. For example: '`\033[40m`. See http://jafrog.com/2013/11/23/colors-in-terminal.html
	for more.
	"""

	_c = dict(
		black='40m',
		red='41m',
		green='42m',
		orange='43m',
		purple='45m',
		blue='46m',
		grey='47m'
	)

	# result will be in this format: [<PREFIX>];[<COLOR>];[<TEXT DECORATION>][<MESSAGE>][<FINISHING SYMBOL>]
	result = '\033[' + _c[color] + message + '\033[0m'

	return print(result)

def link(name, url):
	"""
	Create popover of variable in html.

	Parameters
	----------
	name : :class:`str`
		Name of variable.
	url : :class:`str`
		URL to link to.

	Returns
	-------
	link : :class:`str`
		HTML <a> element.

	Examples
	--------
	>>> link(name=roi, url="#boxplot")

	.. code-block:: html

		<a href="#boxplot" class="anchor">roi</a>
	"""

	_link = '<a href="%s" class="anchor">%s</a>' % (url, name)

	return _link

def popover(name, title, description, url=None, **kwargs):
	"""
	Create popover of variable in html.

	Parameters
	----------
	name : :class:`str`
		Local name of variable.
	title : :class:`str`
		Title of variable in popover.
	description : :class:`str`
		Description of variable in popover.
	url : :class:`str` or :obj:`str`
		URL to include. Default `None`.
	**kwargs : :obj:`str` or :obj:`None`, optional
		Additional properties, relevent for specific content types. Here's a list of available properties:

		.. list-table::
			:class: kwargs
			:widths: 25 50
			:header-rows: 1
			
			* - Property
			  - Description
			* - img : :obj:`dict` {'src': :obj:`str`,'className': :obj:`str`}
			  - Create an image to be included in the popover

	Returns
	-------
	link : :class:`str`
		HTML <a> element.

	Examples
	--------
	>>> description = "Chambers, J., Hastie, T. (1992). Statistical Models in S. Wadsworth & Brooks/Cole."
	>>> popover(name="anova", title="Statistical Models in S", description=description)

	.. code-block:: html 
		
		<a tabindex="0" class="popover-anchor" link-id="anova" data-toggle="popover" data-content="Chambers, J., Hastie, T. (1992). \
		Statistical Models in S. Wadsworth & Brooks/Cole." title="" data-original-title="Statistical Models in S">Chambers, Hastie, 1992</a>
	"""
	# img
	img = kwargs['img'] if 'img' in kwargs else None

	# if image
	if img is not None:
		_img = '<img class="%s" src="%s"></img>'%(img['class'], img['src'])
	else:
		_img = ''

	# if not url
	if url is not None:
		url = '<a class="anchor" href="%s">%s</a>' %(url, url)
		description = '%s<br>%s' % (description, url)
	else:
		description = ''

	# create link
	_link = '<a tabindex="0" class="popover-anchor" link-id="%s" data-toggle="popover" data-content="%s" title="" data-original-title="%s">%s</a>%s'\
		%(name, description, title, title, _img)
		
	# format
	_link = re.sub(r'\s+', ' ', link).strip()

	return _link

def stn(source):
	"""
	Convert to scientific notation.

	Parameters
	----------
	source : :class:`float`
		Original number.

	Returns
	-------
	output : :class:`str`
		Number in scientific notation, if (number < 0.0001) or (if number > 9999).
	"""
	# convert to float
	source = float(source)

	# if less than 0.0001, use scientific notation
	if (source < 0.0001):
		output = '%.2E' % source
	# else if greater than 9999, use scientific notation
	elif (source > 9999):
		output = '%.2E' % source
	# else round number to 4th digit
	else:
		output = '%s' % (round(source, 4))

	return output

def library(required=None):
	"""
	Check if required libraries are available.

	Parameters
	----------
	required : :class:`list`
		List of required libraries.
	"""

	import platform
	import importlib
	import pkg_resources
	import pip
	from distutils.version import StrictVersion
	from pip._internal import main as _main
	
	# start
	console('settings.library()', 'green')

	# for timestamp
	_t0 = datetime.now()
	_f = debug(message='t', source="timestamp")

	# if required is string
	if isinstance(required, str):
		required = [required]

	# packages to download from outside of pypi
	not_pypi = {
		'nslr': "https://gitlab.com/risoms/nslr"
	}

	# try installing and/or importing packages
	try:
		import subprocess, sys
		pipv = pkg_resources.get_distribution("pip").version
		if StrictVersion(pipv) > StrictVersion('10.0.0'):
			# for required packages check if package exists on device
			for package in required:
					# if missing, install
					if importlib.util.find_spec(package) is None:
						# if package is not in PyPi
						if package in not_pypi:
							_package = 'git+%s'%(not_pypi[package])
							#_main(['install', _path])
							subprocess.check_call([sys.executable, '-m', 'pip', 'install', _package])
						# if package is cv2
						elif package == 'cv2':
							#_main(['install', 'opencv-python'])
							subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'opencv-python'])
						# install from PyPi
						else:
							#_main(['install', package])
							subprocess.check_call([sys.executable, '-m', 'pip', 'install', package])
	except Exception as error:
		return error

	# ----finished
	console('%s finished in %s msec' % (_f, ((datetime.now()-_t0).total_seconds()*1000)), 'blue')

def debug(message, source='debug'):
	"""
	Get function or class for console print.

	Parameters
	----------
	message : :class:`str`
		Log message.
	source : :class:`str`
		Origin of call. Either debug or timestamp.

	"""
	import inspect

	#start
	if source == 'debug':
		caller = inspect.getframeinfo(inspect.stack()[1][0])
		event = "%s, line %d, in processing.%s(), %s" % (caller.filename, caller.lineno, caller.function, message)
		print(event)
	elif source == 'timestamp':
		caller = inspect.getframeinfo(inspect.stack()[1][0])
		event = "processing.%s()" % (caller.function)

	return event

def pydoc(path=None, source=None, build=None, copy=False):
	"""
	Generate pydoc docmentation.

	Parameters
	----------
	source : : class: `str`, optional
		pydoc source path, by default None
	build : : class: `str`, optional
	pydoc source path,, by default None
	copy : : class: `str` or bool, optional
		[description], by default False

	Attributes
	----------
	path: : class: `str`
		Location to save the pydoc to.
	"""

	import subprocess

	# for timestamp
	_t0 = datetime.now()
	_f = debug(message='t', source="timestamp")

	# create pydoc
	# if osx
	if sys.platform == "darwin":
		print(source); print(build)
		subprocess.call(['sphinx-build "%s" "%s"' % (source, build)],stdout=subprocess.PIPE, shell=True)
	# if windows
	else:
		# set location of makefile
		# sys.path.append('/Users/mdl-admin/Desktop/R33-analysis-master/output/docs/')

		_file = path + 'make.bat'
		p = subprocess.Popen([_file], cwd=path, shell=False)
		p.communicate()
		os.system("C:\\Windows\\System32\\cmd.exe /c %s" % (path))

	print('created pydoc at ' + build)
	print(console('%s finished in %s msec' %(_f, ((datetime.now()-_t0).total_seconds()*1000))))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
| @purpose: Process participants metadata for analysis and export.  
| @date: Created on Sat May 1 15:12:38 2019  
| @author: Semeon Risom  
| @email: semeon.risom@gmail.com  
| @url: https://semeon.io/d/R33-analysis  
"""

# available classes and functions
__all__ = ['Metadata']

# required external libraries
__required__ = ['pandas','numpy','json']

# core
from pdb import set_trace as breakpoint
import os
import pandas as pd
import numpy as np
import json

# local libraries
if __name__ == '__main__':
	from . import settings

class Metadata():
	"""Process participants metadata for analysis and export."""
	def __init__(self, isLibrary=False):
		"""
		Initiate the mdl.r33.Metadata module.

        Parameters
        ----------
        isLibrary : :obj:`bool`
            Check if required libraries are available. Default `False`.
        """
        #check libraries
		if isLibrary:
			settings.library(__required__)

	@classmethod
	def summary(cls, df, path):
		"""
		Preparing data for use in analysis.

		Parameters
		----------
		df : :obj:`str`
			Pandas dataframe of raw data.
		path : :obj:`str`
			The directory path of the subject data

		Attributes
		----------
		path : :obj:`str`
			Specific directory path used.
		attr2 : :obj:`str`, optional
			Description of `attr2`.

		Returns
		-------
		df : :class:`ndarray`
			Pandas dataframe of processed metadata.

		Notes
		-----    
		You can either get data from all files within a directory (directory), or from a specific
		subject (subject_session).

		Examples
		--------
		>>> #if using path:
		>>> df = getData(path=self.config['path'])

		>>> #if getting data for single subject:
		>>> df = getData(path=self.config['path'],subject_session=['1099','1', '0'])

		"""

		#drop subject 111111, 999999, nan
		df = df.drop(df[(df['participant']==111111)|(df['participant']==999999)].index)
		df['participant'] = df['participant'].apply(pd.to_numeric)

		"""processing data"""
		#rename browser, os, date
		df.rename(columns={'browser':'browser_old','os':'os_old','date':'date_old'}, inplace=True)

		"""gpu_type"""
		df['gpu_type'] = 'integrated'
		df.loc[df['gpu'].str.contains('AMD', na=False),'gpu_type'] = 'dedicated'
		df.loc[df['gpu'].str.contains('Nvidia', na=False),'gpu_type'] = 'dedicated'
		df.loc[df['gpu'].str.contains('NVIDIA', na=False),'gpu_type'] = 'dedicated'
	
		"""webcam brand"""
		df['webcam_brand']  = df['WebcamDevice'].str.split().str.get(0)

		"""webcam width"""
		#replace "-1" with ".x."
		df['webcamSize.px'] = df['webcamSize.px'].apply(lambda x: '.x.' if (x == -1) else x)
		#replace ""0x0"" with ".x."
		df['webcamSize.px'] = df['webcamSize.px'].apply(lambda x: '.x.' if (x == "0x0") else x)
		df[['webcamWidth','webcamHeight']] = df['webcamSize.px'].apply(lambda x: pd.Series([i for i in x.split('x')]))

		"""window width"""
		df[['windowWidth','windowHeight']] = df['windowSize.px'].apply(lambda x: pd.Series([i for i in x.split('x')]))

		"""monitor width"""
		df['monitorSize old'] = df['monitorSize.px']
		#adjust size back to value before manually multiplying by devicePixelratio
		df[['monitorWidth','monitorHeight']] = df['monitorSize.px'].apply(lambda x: pd.Series([i for i in x.split('x')]))
		df['monitorWidth'] = pd.to_numeric(df['monitorWidth']) / df['devicePixelRatio']           
		df['monitorHeight'] = pd.to_numeric(df['monitorHeight']) / df['devicePixelRatio']

		"""browser"""
		df['browser'] = 'None'
		###new column for version without number
		df.loc[df['browser_old'].str.contains('Chrome', na=False),'browser'] = 'Chrome'
		df.loc[df['browser_old'].str.contains('Safari', na=False),'browser'] = 'Safari'
		df.loc[df['browser_old'].str.contains('Edge', na=False),'browser'] = 'Edge'
		df.loc[df['browser_old'].str.contains('Firefox', na=False),'browser'] = 'Firefox'
		df.loc[df['browser_old'].str.contains('IE', na=False),'browser'] = 'IE'

		df['browser_version'] = 'None'
		###new column for version without number
		df['browser_version'] = df['browser_old'].map(lambda x: x.lstrip('Chrome').rstrip('aAbBcC'))
		df['browser_version'] = df['browser_old'].map(lambda x: x.lstrip('Safari').rstrip('aAbBcC'))
		df['browser_version'] = df['browser_old'].map(lambda x: x.lstrip('Edge').rstrip('aAbBcC'))
		df['browser_version'] = df['browser_old'].map(lambda x: x.lstrip('Firefox').rstrip('aAbBcC'))
		df['browser_version'] = df['browser_old'].map(lambda x: x.lstrip('IE').rstrip('aAbBcC'))

		"""os"""
		df['os'] = 'None'
		###new column for version without number
		df.loc[df['os_old'].str.contains('Windows', na=False),'os'] = 'Microsoft Windows'
		df.loc[df['os_old'].str.contains('Mac', na=False),'os'] = 'macOS'
		df.loc[df['os_old'].str.contains('Chrome', na=False),'os'] = 'Chrome OS'
	
		"""os version"""
		df['os_version'] = 'None'
		###new column for version without name
		df['os_version'] = df['os_old'].map(lambda x: x.lstrip('Windows').rstrip('aAbBcC'))
		df['os_version'] = df['os_version'].map(lambda x: x.lstrip('Mac OS X').rstrip('aAbBcC'))  
		df['os_version'] = df['os_version'].map(lambda x: x.lstrip('Chrome OS').rstrip('aAbBcC'))

		"""date"""
		##process date
		df['date'] = [x.strip().replace('_', '-') for x in df['date_old']] ##remove underscore
		df['date'] = pd.to_datetime(df['date']) #convert to pandas datetime format
		df['date'] = df['date'].dt.date #remove time
		df['date'] = df['date'].astype(str)

		#rename
		df = df.rename(columns={'windowSize.px':'windowSize','monitorSize.px':'monitorSize',\
										'webcamSize.px':'webcamSize','lum':'luminance'})

		#convert to cm
		df['diagonalSize.cm'] = df['diagonalSize.in'].map(lambda x: round(x * 2.54, 3))

		"""
		clean unusual resolutions for monitor 
		"""  
		#rev = df['monitorSize'].apply(lambda x: pd.Series([i for i in x.split('x')]))
		#rev.rename(columns={0:'monitorWidth',1:'monitorHeight'},inplace=True)

		#convert to integer
		df['monitorWidth'] = df['monitorWidth'].apply(pd.to_numeric)
		df['monitorHeight'] = df['monitorHeight'].apply(pd.to_numeric)

		#recombine
		df["monitorSize"] = df['monitorWidth'].map(str).str.split('.').str[0] + 'x' + \
							df['monitorHeight'].map(str).str.split('.').str[0]
		"""
		clean unusual dpi   
		"""    
		df['devicePixelRatio'] = df['devicePixelRatio'].apply(pd.to_numeric)

		'''
		modify gpu columns
		'''
		df['gpu'] = df['gpu'].fillna(np.NaN)
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('ANGLE ' , ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('(A' , 'A'))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('(I' , 'I'))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('(N' , 'N'))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('vs_5_' , ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('ps_5_' , ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('vs_3_' , ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('ps_3_' , ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace(' 0 0)' , ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('(R)' , ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('(TM)', ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('OpenGL', ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('Engine', ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('Direct3D11', ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('Direct3D9Ex', ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('Family', ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace('Mesa DRI ', ''))
		df['gpu'] = df['gpu'].map(lambda x: str(x).replace(' (Skylake GT2)', ''))

		"""
		clean white space in column
		"""    
		#remove trailing whitespace in gpu  
		df['gpu'] = df['gpu'].map(lambda x: x.strip())

		'''
		split WebcamDevice to retrieve vendor id and fix formatting
		'''
		##format to np.Nan if x=-1
		df['webcamSize'] = df['webcamSize'].map(lambda x: 
			np.NaN if x=='0x0' else np.where(x==-1, np.NaN, x))

		##format to np.Nan if x=-1
		df['WebcamDevice'] = df['WebcamDevice'].map(lambda x: 
			np.NaN if x==-1 else str(x).replace('(Built-in) ', ''))
		#df['WebcamDeviceVendor'] = df['WebcamDeviceVendor'].map(lambda x: x[x.find("(")+1:x.find(")")])
		df[['WebcamDeviceProductID']] = df['WebcamDevice'].str.split('\(|\)', expand=True).iloc[:,[1]]

		#rename variables
		df = df.rename(columns={'trialNumTask':'TrialNum','Key_Resp.resp':'RT','isWebcamUsed':'is_eyetracking'})
		#drop columns
		cols = ['sampleNum', 'x', 'y', 'duration.t', 'Stim_onset.t', 'DotLoc_onset.t', 'blockNum',\
				'trialNum', 'TrialNum', 'trialID', 'Key_Resp.rt', 'Key_Resp.cresp', 'Key_Resp.acc',\
				'DotLoc', 'LEmotion', 'LStim', 'LDescription', 'REmotion', 'RStim', 'RDescription', 
				'trialType','isCongruent', 'event', 'trial_type', 'internal_node_id', 'RT']
		#drop columns
		# cols = ['sampleNum', 'timestamp', 'x', 'y', 'duration.t', 'Stim_onset.t', 'DotLoc_onset.t', 'blockNum',\
		#          'trialNum', 'TrialNum', 'trialID', 'Key_Resp.rt', 'Key_Resp.resp', 'Key_Resp.cresp', 'Key_Resp.acc',\
		#          'DotLoc', 'LEmotion', 'LStim', 'LDescription', 'REmotion', 'RStim', 'RDescription', 'trialType',\
		#          'isCongruent', 'event', 'trial_type', 'internal_node_id', 'type', 'RT', 'marker', 'bad', 'sg_x',\
		#          'sg_y', 'sg_class', 'left_bound', 'right_bound', 'sg_fix_all', 'sg_fix_index', 'sg_roi_bounds',\
		#          'sg_fix_roi']
		df.drop(cols, inplace=True, axis=1)

		#------------------------------------------------save
		print("demographics saved: %s"%(path))
		df.to_csv(path, index=False)

		return df

	@classmethod
	def predict(cls, df):
		"""
		Predicting screen size (cm), device (i.e. macbook 2018).

		Parameters
		----------
		df : :class:`ndarray`
			Pandas dataframe of raw data.

		Returns
		-------
		df : :class:`ndarray`
			Pandas dataframe of raw data.
		"""
		#clean up sub-version of data
		df['os_version'] = np.where(df['os'] == 'OSX',\
											df['os_version'].map(lambda x: x.replace(r'[^.]+', '')[:-2]),\
											df['os_version']) #else
		"""
		import screensize sample list
		"""
		#import reference screen size
		screensize_path = os.path.abspath(__file__+ '../../../info')
		df_screensize = pd.read_excel(screensize_path+'/screensize.xlsx')
		df_screensize = df_screensize.rename(columns={'resolution (px)': 'monitorSize'}) #rename for merge
		df_screensize = df_screensize.rename(columns={'gpu': 'gpu list'}) #rename for merge

		#remove excel non-breaking space \xa0
		df_screensize['gpu list'] = df_screensize['gpu list'].replace({'\\xa0': ' '}, regex=True) 
		df_screensize['device'] = df_screensize['device'].replace({'\\xa0': ' '}, regex=True) 
		df_screensize['model id'] = df_screensize['model id'].replace({'\\xa0': ' '}, regex=True) 

		#clear leading and trailing white space in string
		df_screensize['gpu list'] = df_screensize['gpu list'].astype(str).map(lambda x: x.strip())

		#convert inches to cm
		df_screensize['screen size (cm)'] = df_screensize['screen size (in)'].map(lambda x: round(x * 2.54, 3))
		tt=df_screensize['gpu list'][0]

		#convert each gpu cell into a list
		#https://stackoverflow.com/a/47548471
		#https://stackoverflow.com/questions/38133961/pandas-how-to-store-a-list-in-a-dataframe
		#https://stackoverflow.com/questions/35565376/insert-list-of-lists-into-single-column-of-pandas-df
		df_screensize['gpu list']=df_screensize['gpu list'].map(lambda x: list(map(str.strip,x.split(","))))

		'''
		merge location (lab or home) df and df_summary data
		'''
		df_all = pd.merge(df, df_summary[['participant','session','location']],on=['participant','session'], how='left')
		df_all.sort_values(by=['participant','session','subsession']).reset_index(drop=True)
		df_lab=df_all.copy().reset_index(drop=True)
	
		'''
		lab computer (if: subject is in lab and using one of the lab machines)
		'''
		#filter
		df_screensize_filter=df_screensize.copy().reset_index(drop=True)
		df_screensize_filter = df_screensize_filter.loc[df_screensize['is lab computer'] == True].reset_index(drop=True)    
		df_screensize_filter['location'] = 'lab'
		df_screensize_filter['exact match'] = True

		#preparing new variables for df_osx
		df_screensize_filter['devices'] = df_screensize_filter['device'] 
		df_screensize_filter['model id'] = df_screensize_filter['model id']
		df_screensize_filter['resolution (px)'] = df_screensize_filter['monitorSize']

		#combine all    
		df_lab = pd.merge(df_lab,df_screensize_filter[['os','gpu list','monitorSize','location','screen size (cm)',\
														'pixel density (ppi)','exact match',\
														'devices','model id', 'resolution (px)']],\
														on=['os','monitorSize','location'], how='left')
		df_lab = df_lab.loc[df_lab['exact match'] == True].reset_index(drop=True) 
		df_lab.sort_values(by=['participant','session','subsession']).reset_index(drop=True)

		#save summary
		types_df_lab = df_lab.dtypes
		df_test = df_lab[['participant','session','subsession','os','os_version','gpu',\
								'diagonalSize.cm','screen size (cm)','gpu list','exact match']]
		df_test.to_excel(cwd_save+'/lab_summary.xlsx', index=False)
	
		'''
		osx devices
		'''
		#filter
		df_screensize_filter=df_screensize.copy().reset_index(drop=True)
		df_screensize_filter = df_screensize.loc[df_screensize['os'] == 'OSX'].reset_index(drop=True)
		df_osx = df_all.loc[df_all['os'] == 'OSX'].reset_index(drop=True)
		df_osx.sort_values(by=['participant','session','subsession']).reset_index(drop=True)

		#check each gpu to see if there is more than one matching
		df_osx['gpu list'] = 'nan' #gpu list
		df_osx['screen size (cm)'] = 'nan' #screen size
		df_osx['pixel density (ppi)'] = 'nan' #pixel density
		df_osx['exact match'] = 'nan' #only single match
		df_osx['devices'] = 'nan' #devices
		df_osx['model id'] = 'nan' #model id
		df_osx['resolution (px)'] = 'nan' #resolution
		for idx, rw in df_osx.iterrows():
			gpu = rw['gpu']
			l_match_d = [] #matching devices
			l_match_mid = [] #matching model id
			l_match_ss = [] #matching screen size
			l_match_res = [] #matching resolution
			l_match_gpu = [] #matching gpu
			l_match_px = [] #matching pixel density
			#for each device
			for index, row in df_screensize_filter.iterrows():
				#l_row = map(str.strip, row['gpu list']) #strip items in list
				l_row = row['gpu list']
				#if gpu in list
				if [x for x in l_row if gpu.lower() in x.lower()].__len__() > 0:
					#add device to list 
					l_match_d.append(str(row['device']))
					l_match_mid.append(str(row['model id']))
					l_match_gpu.append(str(row['gpu list']))
					l_match_px.append(str(row['pixel density (ppi)']))
					l_match_ss.append(str(row['screen size (cm)']))
					l_match_res.append(str(row['monitorSize']))
			
			#if no matches
			if l_match_d.__len__() == 0:
				df_osx['devices'][idx] = 'nan'
				df_osx['model id'][idx] = 'nan'
				df_osx['gpu list'][idx] = 'nan'
				df_osx['pixel density (ppi)'][idx] = 'nan'
				df_osx['screen size (cm)'][idx] = 'nan'
				df_osx['resolution (px)'][idx] = 'nan'
				df_osx['exact match'][idx] = False
	
			#if only one device add immediately            
			elif l_match_d.__len__() == 1:
				df_osx['devices'][idx] = l_match_d[0]
				df_osx['model id'][idx] = l_match_mid[0]
				df_osx['gpu list'][idx] = l_match_gpu[0]
				df_osx['pixel density (ppi)'][idx] = l_match_px[0]
				df_osx['screen size (cm)'][idx] = l_match_ss[0]
				df_osx['resolution (px)'][idx] = l_match_res[0]
				df_osx['exact match'][idx] = True    
		
			#if multiple matches
			elif l_match_d.__len__() > 1:
				#add device to dataframe
				df_osx['devices'][idx] = l_match_d
				df_osx['model id'][idx] = l_match_mid
				df_osx['gpu list'][idx] = l_match_gpu
				df_osx['pixel density (ppi)'][idx] = l_match_px
				df_osx['screen size (cm)'][idx] = l_match_ss
				df_osx['resolution (px)'][idx] = l_match_res
				df_osx['exact match'][idx] = False

		#save summary
		types_df_osx = df_osx.dtypes
		df_test = df_osx[['participant','session','subsession','os','os_version','gpu',\
								'diagonalSize.cm','screen size (cm)','gpu list','exact match']]
		df_test.to_excel(cwd_save+'/osx_summary.xlsx', index=False)

		'''
		chromebook devices
		'''
		#filter
		df_screensize_filter = df_screensize.loc[df_screensize['os'] == 'Chrome'].reset_index(drop=True)
		df_chrome = df_all.loc[df_all['os'] == 'Chrome'].reset_index(drop=True)
		df_chrome.sort_values(by=['participant','session','subsession']).reset_index(drop=True)


		'''#----------------------------attempt 1'''
		#check each gpu to see if there is more than one matching
		df_chrome['gpu list'] = 'nan' #gpu list
		df_chrome['screen size (cm)'] = 'nan' #screen size
		df_chrome['pixel density (ppi)'] = 'nan' #pixel density
		df_chrome['exact match'] = 'nan' #only single match
		df_chrome['devices'] = 'nan' #devices
		df_chrome['model id'] = 'nan' #model id
		df_chrome['resolution (px)'] = 'nan' #resolution
		idx=0
		rw=0
		index=0
		row=0
		for idx, rw in df_chrome.iterrows():
			gpu = rw['gpu']
			l_match_d = [] #matching devices
			l_match_mid = [] #matching model id
			l_match_ss = [] #matching screen size
			l_match_res = [] #matching resolution
			l_match_gpu = [] #matching gpu
			l_match_px = [] #matching pixel density
			#for each device
			for index, row in df_screensize_filter.iterrows():
				#l_row = map(str.strip, row['gpu list']) #strip items in list
				l_row = row['gpu list']
				#if gpu in list
				if [x for x in l_row if gpu.lower() in x.lower()].__len__() > 0:
					#add device to list 
					l_match_d.append(str(row['device']))
					l_match_mid.append(str(row['model id']))
					l_match_gpu.append(str(row['gpu list']))
					l_match_px.append(str(row['pixel density (ppi)']))
					l_match_ss.append(str(row['screen size (cm)']))
					l_match_res.append(str(row['monitorSize']))
			
			#if no matches
			if l_match_d.__len__() == 0:
				df_chrome['devices'][idx] = 'nan'
				df_chrome['model id'][idx] = 'nan'
				df_chrome['gpu list'][idx] = 'nan'
				df_chrome['pixel density (ppi)'][idx] = 'nan'
				df_chrome['screen size (cm)'][idx] = 'nan'
				df_chrome['resolution (px)'][idx] = 'nan'
				df_chrome['exact match'][idx] = False
	
			#if only one device add immediately            
			elif l_match_d.__len__() == 1:
				df_chrome['devices'][idx] = l_match_d[0]
				df_chrome['model id'][idx] = l_match_mid[0]
				df_chrome['gpu list'][idx] = l_match_gpu[0]
				df_chrome['pixel density (ppi)'][idx] = l_match_px[0]
				df_chrome['screen size (cm)'][idx] = l_match_ss[0]
				df_chrome['resolution (px)'][idx] = l_match_res[0]
				df_chrome['exact match'][idx] = True    
		
			#if multiple matches
			elif l_match_d.__len__() > 1:
				#add device to dataframe
				df_chrome['devices'][idx] = l_match_d
				df_chrome['model id'][idx] = l_match_mid
				df_chrome['gpu list'][idx] = l_match_gpu
				df_chrome['pixel density (ppi)'][idx] = l_match_px
				df_chrome['screen size (cm)'][idx] = l_match_ss
				df_chrome['resolution (px)'][idx] = l_match_res
				df_chrome['exact match'][idx] = False

		#save summary
		types_df_chrome = df_chrome.dtypes
		df_test = df_chrome[['participant','session','subsession','os','os_version','gpu',\
								'diagonalSize.cm','screen size (cm)','gpu list','exact match']]
		df_test.to_excel(cwd_save+'/chrome_summary.xlsx', index=False)           


		'''
		combine data---------osx, chrome, and lab
		'''     
		#sort
		df_lab.sort_values(by=['participant','session','subsession']).reset_index(drop=True)
		df_osx.sort_values(by=['participant','session','subsession']).reset_index(drop=True)
		df_chrome.sort_values(by=['participant','session','subsession']).reset_index(drop=True)
		#concat
		df_merge = pd.concat([df_lab, df_osx, df_chrome], ignore_index=True)
		df_merge.sort_values(by=['participant','session','subsession']).reset_index(drop=True)
		#arrange columns
		df_merge = df_merge.reindex(list(df_osx.columns), axis=1)

		'''
		combine data---------all
		'''
		df_merge_small = df_merge[
				['participant','session','subsession','devices','exact match','resolution (px)',
				'screen size (cm)','gpu list', 'pixel density (ppi)']]
		df_f = pd.merge(df, df_merge_small, on=['participant','session','subsession'], how='outer')

		#fix gpu lists
		df_f['gpu list'] = df_f['gpu list'].map(lambda x: str(x).replace('"[' , '['))
		df_f['gpu list'] = df_f['gpu list'].map(lambda x: str(x).replace(']"' , ']'))

		#fix devices lists
		df_f['devices'] = df_f['devices'].astype(str) 
		df_f['resolution (px)'] = df_f['resolution (px)'].astype(str) 
		df_f['screen size (cm)'] = df_f['screen size (cm)'].astype(str) 
		df_f['pixel density (ppi)'] = df_f['pixel density (ppi)'].astype(str) 

		#drop gpu lists
		df_f = df_f[df_f.columns.drop('gpu list')]

		#merge location (lab or home) df and df_summary data
		##prepare
		df_sum = df_summary[['participant','session','subsession','location']]
		df_sum.sort_values(by=['participant','session','subsession']).reset_index(drop=True)
		df_f.sort_values(by=['participant','session','subsession']).reset_index(drop=True)
		##drop unusual participants
		df_f = df_f.drop(df_f[(df_f['participant']==111111) 
							| (df_f['participant']==999999)].index)

		df_sum = df_sum.drop(df_sum[(df_sum['participant']==111111) 
							| (df_sum['participant']==999999)].index)

		##convert types
		df_f['participant'] = df_f['participant'].astype(float)
		df_f['session'] = df_f['session'].astype(float)
		df_sum['participant'] = df_sum['participant'].astype(float)
		df_sum['session'] = df_sum['session'].astype(float)

		##merge
		df_f = pd.merge(df_f, df_sum, 
							on=['participant','session','subsession'], how='left')

		#export
		finished_json = df_f.to_json(orient='records')
		finished_json = 'json_data =' + finished_json
		with open(cwd_save+'/summary.json', 'w+') as f:
			f.write(json.dumps(finished_json,indent=4).strip('"').replace('\\',''))
			f.close()
	
		return df
